from celescope.tools.utils import STAR_util, add_log


@add_log
def mapping_mut(args):
    STAR_util(
        args.sample,
        args.outdir,
        args.fq,
        args.indel_genomeDir,
        args.thread,
        args.outFilterMatchNmin,
        outBAMsortingBinsN=200,
    )


def get_opts_mapping_mut(parser, sub_program):
    if sub_program:
        parser.add_argument('--outdir', help='output dir', required=True)
        parser.add_argument('--sample', help='sample name', required=True)
        parser.add_argument("--fq", required=True)
        parser.add_argument('--assay', help='assay', required=True)
    parser.add_argument(
        '--indel_genomeDir',
        help='insertion or deletion STAR indexed genome directory',
        required=True)
    parser.add_argument("--thread", help='STAR thread', default=1)
    parser.add_argument("--outFilterMatchNmin", help='STAR outFilterMatchNmin', default=35)



#!/usr/bin/env python
# v1.0

import pysam
import os
import subprocess
import numpy as np
import pandas as pd
from celescope.tools.step import Step, s_common
import celescope.tools.utils as utils


class Conversion(Step):
    """
    Features
    - Get conversion pos in each read.
	- Get snp info. 

    Output
    - `{sample}.PosTag.bam` Bam file with conversion info.
    - `{sample}.PosTag.csv` SNP info in csv format.
    """

    def __init__(self, args, step_name):
        Step.__init__(self, args, step_name)
        # input files
        self.ifile = os.path.join(args.outdir, args.sample+'.bam')
        self.sample = args.sample
        self.strandednessfile = args.strand
        self.inbam = args.bam
        self.bcfile = args.cell
        self.outdir = args.outdir
        self.thread = args.thread

        # output files
        self.outfile_bam = os.path.join(args.outdir, args.sample+'.PosTag.bam')
        self.outfile_csv = os.path.join(args.outdir, args.sample+'.PosTag.csv')

    @utils.add_log
    def run(self):
        ##Filter and sort
        self.fltSort(self.inbam,self.ifile,self.bcfile,self.thread)
        cmd=['samtools index',self.ifile]
        self.run_cmd(cmd)

        ##Adding tags
        self.addTags(self.ifile,self.outfile_bam,self.strandednessfile)
        cmd=['samtools index',self.outfile_bam]
        self.run_cmd(cmd)

        #Obtaining conversion positions
        bam = pysam.AlignmentFile(self.outfile_bam, 'rb')
        ContigLocs, AnnoteLocs=self.CountConvperPos(bam)

        #Obtaining coverage over conversion position   
        ConvsPerPos,CoverofPosWithConvs = self.CountReadConverPerConvPos(bam,ContigLocs)
        A=self.ExportasVcf(ConvsPerPos,CoverofPosWithConvs,AnnoteLocs)
        A['sample']  = self.sample
        #Saving result
        A.to_csv(self.outfile_csv)
        bam.close()

        cmd=['rm', self.ifile]
        self.run_cmd(cmd)
        cmd=['rm', self.ifile+'.bai']
        self.run_cmd(cmd)

    def run_cmd(self,cmd):
        subprocess.call(' '.join(cmd),shell=True)

    @utils.add_log
    def CountConvperPos(self,bamfile):
        ContigLocs={}
        AnnoteLocs={}
        for read in bamfile.fetch():
            try:
                if read.get_tag('ST')=='+':
                    locs=read.get_tag('TL')
                else:
                    locs=read.get_tag('AL')
                if locs[0]!=0:
                    if read.reference_name in ContigLocs:
                        ContigLocs[read.reference_name].extend(locs)
                    else:
                        ContigLocs[read.reference_name] = list(locs)
                    if read.reference_name not in AnnoteLocs:
                        for i,each in enumerate(locs):
                            if i == 0:
                                AnnoteLocs[read.reference_name] = { each :read.get_tag('XT')}
                            else:
                                AnnoteLocs[read.reference_name][each] = read.get_tag('XT')
                    else:
                        for i,each in enumerate(locs):
                            if each not in AnnoteLocs[read.reference_name]:
                                AnnoteLocs[read.reference_name][each] = read.get_tag('XT')
            except (ValueError,KeyError):
                continue
        return ContigLocs, AnnoteLocs

    @utils.add_log
    def CountReadConverPerConvPos(self,bam,ContigLocs):
        ConvsPerPos={}
        CoverofPosWithConvs={}
        for key in ContigLocs.keys():
            ContigLocs[key]=sorted(ContigLocs[key])
            ConvsPerPos[key]={}
            k=0
            current=ContigLocs[key][k]
            k+=1
            nextone=ContigLocs[key][k]
            while k < len(ContigLocs[key])-1:
                ConvsPerPos[key][current]=1
                while current == nextone and k < len(ContigLocs[key])-1:
                    k+=1
                    nextone=ContigLocs[key][k]
                    ConvsPerPos[key][current]+=1
                current = nextone
                if k < len(ContigLocs[key])-1:
                    k+=1
                    nextone=ContigLocs[key][k]

            CoverofPosWithConvs[key]={}
            for key2 in ConvsPerPos[key].keys():
                try:
                    CoverofPosWithConvs[key][key2]=bam.count(key,key2,key2+1)
                except ValueError:
                    continue
        return ConvsPerPos,CoverofPosWithConvs

    @utils.add_log
    def ExportasVcf(self,ConvsPerPos,CoverofPosWithConvs, AnnoteLocs):
        #Chrom, Pos , ConvsPerPs, CoverofPosWithConvs
        Outputdf =pd.DataFrame(columns=['pos2','convs','covers','chrom','posratio'])
        for key in ConvsPerPos.keys():
            df=pd.DataFrame.from_dict(ConvsPerPos[key], orient='index')
            df1=pd.DataFrame.from_dict(CoverofPosWithConvs[key], orient='index')
            df.index.name='pos'
            df1.index.name='pos'
            df.columns = ['convs']
            df1.columns = ['covers']
            df2=df.join(df1)
            df2['pos2'] = df2.index
            df2.index = np.arange(df2.shape[0])
            df2['chrom']=np.repeat(key,df2.shape[0])
            df2['posratio']=df2['convs']/df2['covers']
            df3=pd.DataFrame.from_dict(AnnoteLocs[key], orient='index')
            df3.columns = ['gene_id']
            df2=df2.join(df3, on='pos2')
            Outputdf=Outputdf.append(df2)
        return Outputdf.reset_index(drop=True)

    def createTag(self,d):
        return ''.join([''.join(key) + str(d[key]) + ';' for key in d.keys()])[:-1]


    def convInRead(self, read, qual = 20):
        specific_conversions = {}
        total_content = {'a' : 0, 'c' : 0, 'g' : 0, 't' : 0}
        specific_conversions[('c', 'A')] = 0
        specific_conversions[('g', 'A')] = 0
        specific_conversions[('t', 'A')] = 0
        specific_conversions[('a', 'C')] = 0
        specific_conversions[('g', 'C')] = 0
        specific_conversions[('t', 'C')] = 0
        specific_conversions[('a', 'G')] = 0
        specific_conversions[('c', 'G')] = 0
        specific_conversions[('t', 'G')] = 0
        specific_conversions[('a', 'T')] = 0
        specific_conversions[('c', 'T')] = 0
        specific_conversions[('g', 'T')] = 0
        specific_conversions[('a', 'N')] = 0
        specific_conversions[('c', 'N')] = 0
        specific_conversions[('g', 'N')] = 0
        specific_conversions[('t', 'N')] = 0

        tC_loc = []
        aG_loc = []

        try:
            refseq = read.get_reference_sequence().lower()
        except (UnicodeDecodeError):
            refseq=''

        for base in total_content.keys():
            total_content[base] += refseq.count(base)
        for pair in read.get_aligned_pairs(with_seq=True):
            try:
                if pair[0] is not None and pair[1] is not None and pair[2] is not None:
                    if str(pair[2]).islower() and not read.query_qualities[pair[0]] < qual:
                        specific_conversions[(pair[2],read.seq[pair[0]])] += 1
                        if (pair[2],read.seq[pair[0]]) == ('t', 'C'):
                            tC_loc.append(pair[1])
                        if (pair[2],read.seq[pair[0]]) == ('a', 'G'):
                            aG_loc.append(pair[1])
            except (UnicodeDecodeError, KeyError):
                continue
        SC_tag = self.createTag(specific_conversions)
        TC_tag = self.createTag(total_content)
        
        if len(tC_loc) == 0:
            tC_loc.append(0)
        if len(aG_loc) == 0:
            aG_loc.append(0)
        return SC_tag, TC_tag, tC_loc, aG_loc

    @utils.add_log
    def addTags(self,bamfilename, outputname,strandednessfile):
        bamfile = pysam.AlignmentFile(bamfilename, 'rb')
        mod_bamfile = pysam.AlignmentFile(outputname, mode='wb',template=bamfile)
        strandedness = pd.read_csv(strandednessfile, header=None, index_col=0)
        for read in bamfile.fetch():
            try:
                tags = self.convInRead(read)
                read.set_tag('SC',tags[0],'Z')
                read.set_tag('TC',tags[1],'Z')
                read.set_tag('TL',tags[2])
                read.set_tag('AL',tags[3])
                read.set_tag('ST',strandedness.loc[read.get_tag('XT')][1])
                mod_bamfile.write(read)
            except (ValueError,KeyError):
                continue

        bamfile.close()
        mod_bamfile.close()

    @utils.add_log
    def fltSort(self,bamfilename, outfile_bam,cellfile, thread=8):
        bamfile = pysam.AlignmentFile(bamfilename, 'rb')
        mod_bamfile = pysam.AlignmentFile(outfile_bam, mode='wb',template=bamfile)
        cells={}
        with open(cellfile) as f:
            for i in f:
                cells[i.strip()] = 1
        for read in bamfile.fetch(until_eof=True):
            try:
                if not read.has_tag('GX'): continue
                if read.get_tag("CB") not in cells: continue
                mod_bamfile.write(read)
            except (ValueError,KeyError):
                continue
        bamfile.close()
        mod_bamfile.close()

        cmd=['samtools sort -@',str(thread), '-o', outfile_bam+'.bam',outfile_bam]
        self.run_cmd(cmd)
        cmd=['mv',outfile_bam+'.bam',outfile_bam]
        self.run_cmd(cmd)





@utils.add_log
def conversion(args):

    step_name = "conversion"
    conversion_obj = Conversion(args, step_name)
    conversion_obj.run()

def get_opts_conversion(parser, sub_program):
    parser.add_argument('--strand', help='gene strand file, the format is "geneID,+/-"', required=True)
    if sub_program:
        parser.add_argument("--bam", help='featureCount bam(sortedByCoord), must have "MD" tag, set in star step', required=True)
        parser.add_argument("--cell", help='barcode cell list', required=True)
        parser = s_common(parser)    
    return parser




#!/bin/env python
# coding=utf8

import os
import pysam
import re
import pandas as pd
import plotly
import plotly.graph_objects as go
from celescope.tools.step import Step, s_common
import celescope.tools.utils as utils


class Subsitution(Step):
    """
    Features
    - Computes the overall conversion rates in reads and plots a barplot.

    Output
    - `{sample}.substitution.txt` Tab-separated table of the overall conversion rates.
    """

    def __init__(self, args, step_name):
        Step.__init__(self, args, step_name)

        # input files
        self.sample = args.sample 
        self.bam_file = args.bam
        self.outdir = args.outdir

        # output files
        self.outstat = os.path.join(self.outdir, self.sample+'.substitution.txt')

    
    @utils.add_log
    def run(self):
        # overall rate
        for_base,rev_base,is_forward,is_reverse = self.get_sub_tag(self.bam_file)
        self.sub_stat(for_base,rev_base,is_forward,is_reverse,self.outstat)
        div_item = self.sub_plot(self.outstat)

        self.report_prepare(div_item)
        self.clean_up()


    @utils.add_log
    def get_sub_tag(self,bam):
        bamfile = pysam.AlignmentFile(bam, 'rb')
        is_reverse = {'cA':0, 'gA':0, 'tA':0, 'aC':0, 'gC':0, 'tC':0, 'aG':0, 'cG':0, 'tG':0, 'aT':0, 'cT':0, 'gT':0}
        is_forward = {'cA':0, 'gA':0, 'tA':0, 'aC':0, 'gC':0, 'tC':0, 'aG':0, 'cG':0, 'tG':0, 'aT':0, 'cT':0, 'gT':0}
        for_base = {'a':0, 'c':0, 'g':0, 't':0}
        rev_base = {'a':0, 'c':0, 'g':0, 't':0}
        snp_tags = ['','cA', 'gA', 'tA', 'aC', 'gC', 'tC', 'aG', 'cG', 'tG', 'aT', 'cT', 'gT']
        ref_tags = ['','a','c','g','t']
        for read in bamfile.fetch():
            try:
                snpmatch = re.match( r'cA(\d+);gA(\d+);tA(\d+);aC(\d+);gC(\d+);tC(\d+);aG(\d+);cG(\d+);tG(\d+);aT(\d+);cT(\d+);gT(\d+);', read.get_tag('SC'), re.M)
                totmatch = re.match( r'a(\d+);c(\d+);g(\d+);t(\d+)', read.get_tag('TC'), re.M)
                if snpmatch and totmatch:
                    if read.is_reverse:
                        for j in range(1,len(ref_tags)):
                            rev_base[ref_tags[j]] += int(totmatch.group(j))
                        for i in range(1,len(snp_tags)):
                            is_reverse[snp_tags[i]] += int(snpmatch.group(i))
                    else:
                        for j in range(1,len(ref_tags)):
                            for_base[ref_tags[j]] += int(totmatch.group(j))
                        for i in range(1,len(snp_tags)):
                            is_forward[snp_tags[i]] += int(snpmatch.group(i))
            except (ValueError,KeyError):
                continue
        bamfile.close()

        return for_base,rev_base,is_forward,is_reverse

    @utils.add_log
    def sub_stat(self,for_base,rev_base,is_forward,is_reverse,outfile):
        convertdict = {'a':['aC','aG','aT'],
                    'c':['cA','cG','cT'],
                    'g':['gA','gC','gT'],
                    't':['tA','tC','tG']}
        subdict = {'a':'t','t':'a','c':'g','g':'c',
                'aC':'tG','aG':'tC','aT':'tA',
                'cA':'gT','cG':'gC','cT':'gA',
                'gA':'cT','gC':'cG','gT':'cA',
                'tA':'aT','tC':'aG','tG':'aC'}
        outdict = {'aC':'A_to_C','aG':'A_to_G','aT':'A_to_T',
                'cA':'C_to_A','cG':'C_to_G','cT':'C_to_T',
                'gA':'G_to_A','gC':'G_to_C','gT':'G_to_T',
                'tA':'T_to_A','tC':'T_to_C','tG':'T_to_G'}
        outw = open(outfile,'w')
        for x in ['a','c','g','t']:
            fbase = for_base[x]
            rbase = rev_base[subdict[x]]
            for y in convertdict[x]:
                fcov = is_forward[y]*100 / float(fbase)
                rcov = is_reverse[subdict[y]]*100 / float(rbase)
                outw.write(outdict[y]+'\t'+"%.3f"%fcov+'\t'+"%.3f"%rcov+'\n')
        outw.close()

    @utils.add_log
    def sub_plot(self,txt):
        df = pd.read_table(txt, header=None)
        df.columns = ['sample', '+', '-']

        fig = go.Figure()
        ## 设置颜色：
        import plotly.express as px
        num4colors  = 0
        num4rainbow = 0
        colors_list = []
        while num4colors<100:
            if num4rainbow == 9:
                num4rainbow = 0
            colors_list.append(px.colors.qualitative.Plotly[num4rainbow])
            num4colors+=1
            num4rainbow+=1

        num4sample = 0
        colors4sample = {}
        num4x = 0

        for sample in df['sample'].unique():    
            legend_show = True
            colors4sample[sample] = colors_list[num4sample]
            num4sample += 1
            flag_x = 'x' + str(num4x+1)
            df_plot = df[ df['sample'] == sample ]
            num4x+=1
        
            fig.add_trace(go.Bar(name=sample+'+',  
                    x=df_plot['sample'], 
                    y=df_plot['+'],
                    legendgroup=sample,
                    marker_color=colors4sample[sample],
                    marker_line_color='#FFFFFF',
                    showlegend=legend_show,
                    xaxis=flag_x)
            )
            fig.add_trace(go.Bar(name=sample+'-',  
                    x=df_plot['sample'], 
                    y=df_plot['-'],
                    legendgroup=sample,
                    showlegend=legend_show,
                    marker_color=colors4sample[sample],
                    marker_line_color='#FFFFFF',
                    opacity=0.3,
                    xaxis=flag_x)
            )

        fig.update_layout(barmode='stack')

        per     = 1/(num4x+1)
        gap4bar = per/len(df['sample'].unique())
        num4x = 0
        for typeB in df['sample'].unique():
            if num4x == 0:
                flag_x = 'xaxis'
            else:
                flag_x = 'xaxis' + str(num4x+1)
            anchor_x = 'x'+str(num4x+1)
            num4x += 1
            fig['layout'][flag_x] = dict(domain=[per*num4x, per*(num4x+1)-gap4bar], anchor=anchor_x, title=typeB)

        fig.update_layout(plot_bgcolor = '#FFFFFF')
        fig.update_xaxes(showgrid=False, linecolor='black', showline=True, ticks='outside', showticklabels=False)
        fig.update_yaxes(showgrid=False, linecolor='black', showline=True, ticks='outside')
        width_num = 400 * ( len(df['sample'].unique())* len(df['sample'].unique()) ) / (5*12) ## 控制柱形图的宽度
        fig.update_layout(height=500, width=width_num)
        fig.update_layout(legend=dict(orientation="h"))
        fig.update_layout(legend=dict(
            yanchor="top",
            y=1.3,
            xanchor="left",
            x=0.05,
            valign="top",
        ))

        fig.update_layout(
        yaxis_title="Rates of nucleotide substitution (%)",
        )
        fig.update_xaxes(
            tickangle = -80,
            title_font = {"size": 15},
            title_standoff = 25
        )

        div = plotly.offline.plot(fig, include_plotlyjs=False, output_type='div')

        return div


    def report_prepare(self,outdiv):
        self.add_data_item(substitution=outdiv)


@utils.add_log
def substitution(args):

    step_name = "substitution"
    substitution_obj = Subsitution(args, step_name)
    substitution_obj.run()

def get_opts_substitution(parser, sub_program):
    if sub_program:
        parser.add_argument('--bam', help='bam file from conversion step', required=True)
        parser = s_common(parser) 
    return parser




from collections import defaultdict
from itertools import groupby

import pandas as pd
import pysam

import celescope.tools.utils as utils
from celescope.tools.count import Count, get_opts_count


class Count_capture_rna(Count):

    def bam2table(self):
        """
        read probe file
        """
        probe_gene_count_dict = utils.genDict(dim=4, valType=int)

        samfile = pysam.AlignmentFile(self.bam, "rb")
        with open(self.count_detail_file, 'wt') as fh1:
            fh1.write('\t'.join(['Barcode', 'geneID', 'UMI', 'count']) + '\n')

            def keyfunc(x): return x.query_name.split('_', 1)[0]
            for _, g in groupby(samfile, keyfunc):
                gene_umi_dict = defaultdict(lambda: defaultdict(int))
                for seg in g:
                    (barcode, umi, probe) = seg.query_name.split('_')[:3]
                    if probe != 'None':
                        probe_gene_count_dict[probe]['total'][barcode][umi] += 1
                        if seg.has_tag('XT'):
                            geneID = seg.get_tag('XT')
                            geneName = self.id_name[geneID]
                            probe_gene_count_dict[probe][geneName][barcode][umi] += 1
                        else:
                            probe_gene_count_dict[probe]['None'][barcode][umi] += 1
                    if not seg.has_tag('XT'):
                        continue
                    geneID = seg.get_tag('XT')
                    gene_umi_dict[geneID][umi] += 1
                for gene_id in gene_umi_dict:
                    Count.correct_umi(gene_umi_dict[gene_id])

                # output
                for gene_id in gene_umi_dict:
                    for umi in gene_umi_dict[gene_id]:
                        fh1.write('%s\t%s\t%s\t%s\n' % (barcode, gene_id, umi,
                                                        gene_umi_dict[gene_id][umi]))

        # out probe
        row_list = []
        for probe in probe_gene_count_dict:
            for geneName in probe_gene_count_dict[probe]:
                barcode_count = len(probe_gene_count_dict[probe][geneName])
                umi_count = 0
                read_count = 0
                for barcode in probe_gene_count_dict[probe][geneName]:
                    for umi in probe_gene_count_dict[probe][geneName][barcode]:
                        umi_count += len(probe_gene_count_dict[probe][geneName][barcode])
                        read_count += probe_gene_count_dict[probe][geneName][barcode][umi]
                row_list.append({
                    'probe': probe,
                    'gene': geneName,
                    'barcode_count': barcode_count,
                    'read_count': read_count,
                    'UMI_count': umi_count
                })

        df_probe = pd.DataFrame(row_list,
                                columns=['probe', 'gene', 'barcode_count', 'read_count', 'UMI_count'])
        df_probe = df_probe.groupby(['probe']).apply(
            lambda x: x.sort_values('UMI_count', ascending=False)
        )
        return df_probe

    def run(self):
        df_probe = self.bam2table()
        df_probe.to_csv(f'{self.outdir}/{self.sample}_probe_gene_count.tsv', sep='\t', index=False)

        df = pd.read_table(self.count_detail_file, header=0)
        # df_sum
        df_sum = Count.get_df_sum(df)

        # export all matrix
        self.write_matrix_10X(df, self.raw_matrix_10X_dir)

        # call cells
        cell_bc, _threshold = self.cell_calling(df_sum)

        # get cell stats
        CB_describe = self.get_cell_stats(df_sum, cell_bc)

        # export cell matrix
        df_cell = df.loc[df['Barcode'].isin(cell_bc), :]
        self.write_matrix_10X(df_cell, self.cell_matrix_10X_dir)
        (CB_total_Genes, CB_reads_count, reads_mapped_to_transcriptome) = self.cell_summary(
            df, cell_bc)

        # downsampling
        cell_bc = set(cell_bc)
        saturation, res_dict = self.downsample(df_cell)

        # summary
        self.get_summary(saturation, CB_describe, CB_total_Genes,
                         CB_reads_count, reads_mapped_to_transcriptome)

        self.report_prepare()

        self.add_content_item('metric', downsample_summary=res_dict)
        self.clean_up()


@utils.add_log
def count_capture_rna(args):
    # TODO!
    # need barcode_capture_rna
    step_name = "count_capture_rna"
    runner = Count_capture_rna(args, step_name)
    runner.run()


def get_opts_count_capture_rna(parser, sub_program):
    get_opts_count(parser, sub_program)



import os

from celescope.tools.utils import parse_match_dir

CITESEQ_DIR = os.path.dirname(__file__)


def get_opts_analysis_cite(parser, sub_program):
    if sub_program:
        parser.add_argument('--outdir', help='output dir', required=True)
        parser.add_argument('--sample', help='sample name', required=True)
        parser.add_argument('--match_dir', help='match_dir', required=True)
        parser.add_argument('--citeseq_mtx', help='citeseq matrix .gz file', required=True)
        parser.add_argument('--assay', help='assay', required=True)


def analysis_cite(args):

    if not os.path.exists(args.outdir):
        os.system('mkdir -p %s' % args.outdir)

    rds = parse_match_dir(args.match_dir)['rds']
    app = CITESEQ_DIR + "/analysis_cite.R"
    cmd = (
        f'Rscript {app} '
        f'--rds {rds} '
        f'--citeseq_mtx {args.citeseq_mtx} '
        f'--outdir {args.outdir} '
        f'--sample {args.sample} '
    )
    os.system(cmd)



import pandas as pd


class Barcode_index():
    def __init__(self, barcodes):
        self.barcodes = barcodes
        cell_index = 0
        index_dict = {}
        for barcode in self.barcodes:
            cell_index += 1
            index_dict[barcode] = cell_index
        self.index_dict = index_dict
        self.df_index = pd.DataFrame.from_dict(index_dict, orient='index')
        self.df_index.columns = ['cell_index']
        self.df_index.index.name = 'barcode'

    @staticmethod
    def read_index(index_file):
        df_index = pd.read_csv(index_file, sep='\t', index_col=0, dtype=object)
        df_valid = df_index[df_index['valid'] == 'True']
        return df_valid

    def write_index(self, file_name):
        """
        write index-barcode to file
        """
        self.df_index.to_csv(file_name, sep='\t')



__ASSAY__ = 'capture_rna'

__STEPS__ = [
    'sample',
    'barcode',
    'cutadapt',
    'star',
    "featureCounts",
    "count_capture_rna",
    'analysis']

IMPORT_DICT = {
    'star': 'celescope.rna',
    'analysis': 'celescope.rna',
}



import pandas as pd

import celescope.tools.utils as utils
from celescope.tools.analysis_mixin import AnalysisMixin
from celescope.tools.step import Step, s_common


class Analysis_tag(Step, AnalysisMixin):
    """
    Features
    - Combine scRNA-Seq clustering infromation with tag assignment.
    """

    def __init__(self, args, step_name):
        Step.__init__(self, args, step_name)
        AnalysisMixin.__init__(self, args)

    def run(self):
        self.run_analysis()
        tsne_tag_df = pd.read_csv(self.args.tsne_tag_file, sep="\t", index_col=0)
        feature_tsne = self.get_cluster_tsne(colname='tag', tsne_df=tsne_tag_df, show_colname=False)
        self.add_data_item(cluster_tsne=self.cluster_tsne)
        self.add_data_item(feature_tsne=feature_tsne)
        self.add_data_item(table_dict=self.table_dict)

        self.clean_up()


def get_opts_analysis_tag(parser, sub_program):
    if sub_program:
        parser.add_argument('--tsne_tag_file', help='`{sample}_tsne_tag.tsv` from count_tag. ', required=True)
        parser.add_argument("--match_dir", help="Match celescope scRNA-Seq directory. ")
        parser.add_argument("--tsne_file", help="t-SNE coord file.")
        parser.add_argument("--marker_file", help="Marker file.")
        parser = s_common(parser)


@utils.add_log
def analysis_tag(args):
    step_name = 'analysis_tag'
    ana = Analysis_tag(args, step_name)
    ana.run()



from celescope.fusion.__init__ import __ASSAY__
from celescope.tools.multi import Multi


class Multi_fusion(Multi):
    """
    Features
    - Generate multi-sample scripts.

    Usage
    ```
    multi_fusion\\
    --mapfile ./fusion.mapfile\\
    --fusion_genomeDir {fusion_genomeDir}\\  
    --mod shell
    ```
    """

    def star_fusion(self, sample):
        step = 'star_fusion'
        cmd_line = self.get_cmd_line(step, sample)
        fq = f'{self.outdir_dic[sample]["cutadapt"]}/{sample}_clean_2.fq{self.fq_suffix}'
        cmd = (
            f'{cmd_line} '
            f'--fq {fq} '
        )
        self.process_cmd(cmd, step, sample, m=self.args.starMem, x=self.args.thread)

    def count_fusion(self, sample):
        step = 'count_fusion'
        cmd_line = self.get_cmd_line(step, sample)
        bam = f'{self.outdir_dic[sample]["star_fusion"]}/{sample}_Aligned.sortedByCoord.out.bam'
        cmd = (
            f'{cmd_line} '
            f'--bam {bam} '
            f'--match_dir {self.col4_dict[sample]} '
        )
        self.process_cmd(cmd, step, sample, m=15, x=1)


def main():
    multi = Multi_fusion(__ASSAY__)
    multi.run()


if __name__ == '__main__':
    main()



import configparser
import os
import subprocess

import celescope.tools.utils as utils
from celescope.tools.mkref import Mkref
from celescope.tools.mkref import get_opts_mkref as opts
from celescope.__init__ import HELP_DICT


class Mkref_snp(Mkref):
    """
    Features
    - Create dictionary file and fasta index for gatk SplitNCigarReads.
    (https://gatk.broadinstitute.org/hc/en-us/articles/360035531652-FASTA-Reference-genome-format) 
    Need to run `celescope rna mkref` first

    Output
    - fasta index
    - gatk dictionary file

    Usage
    ```
    # run celescope rna mkref first
    celescope snp mkref \\
     --genome_name Homo_sapiens_ensembl_99 \\
     --fasta Homo_sapiens.GRCh38.dna.primary_assembly.fa
    ```
    """

    def __init__(self, genome_type, args):
        Mkref.__init__(self, genome_type, args)
        self.fasta = args.fasta

    @utils.add_log
    def write_config(self):
        config = configparser.ConfigParser()
        config.read(self.config_file)
        genome = config['genome']
        genome['fasta_gatk_index'] = str(True)
        with open(self.config_file, 'w') as config_handle:
            config.write(config_handle)

    @utils.add_log
    def build_fasta_index(self):
        cmd = (
            f'samtools faidx {self.fasta}'
        )
        Mkref_snp.build_fasta_index.logger.info(cmd)
        subprocess.check_call(cmd, shell=True)

    @utils.add_log
    def build_fasta_dict(self):
        cmd = (
            f'gatk CreateSequenceDictionary -R {self.fasta}'
        )
        Mkref_snp.build_fasta_dict.logger.info(cmd)
        subprocess.check_call(cmd, shell=True)

    @utils.add_log
    def run(self):
        os.chdir(self.genomeDir)
        assert os.path.exists(self.config_file), (
            f"{self.config_file} not exist! "
            "You need to build rna genome with 'celescope rna mkref' first."
        )
        if not self.dry_run:
            self.build_fasta_index()
            self.build_fasta_dict()
        self.write_config()


def mkref(args):
    genome_type = 'snp'
    runner = Mkref_snp(genome_type, args)
    runner.run()


def get_opts_mkref(parser, sub_program):
    opts(parser, sub_program)
    if sub_program:
        parser.add_argument("--fasta", help=HELP_DICT['fasta'], required=True)



#!/usr/bin/env python
#
# Copyright (c) 2015 10X Genomics, Inc. All rights reserved.
#

import sys

import numpy as np
import scipy.stats as sp_stats

# import cellranger.constants as cr_constants
# import tenkit.constants as tk_constants
# import tenkit.seq as tk_seq
# import tenkit.stats as tk_stats

# def to_col_vec(a):
#     """ Convert a 1-d array to a column vector """
#     return np.reshape(a, (len(a), 1))

# def create_gmm(weights, means, sd):
#     """ Create a 2-component GMM with tied variance and given initialization
#         This uses the sklearn 0.17.1 interface - it changes in 0.18.x """
#     gmm = sk_mix.GMM(n_components=2,
#                      covariance_type='tied',
#                      init_params='',
#                      params='wmc')
#     gmm.weights_ = np.array(weights)
#     gmm.means_ = np.reshape(means, (len(means), 1))
#     gmm._set_covars(np.reshape(sd, (1,1)))
#     return gmm


# def multistart_gmm(data, weights, means_list, sd):
#     """ Sweep over the given initial mean vectors
#         and return the result with the highest log-likelihood """
#     best_gmm = None
#     max_loglk = float('-inf')
#     for means in means_list:
#         gmm = create_gmm(weights=weights, means=means, sd=sd)
#         gmm.fit(data)

#         # sklearn 0.17 return type
#         loglk = np.sum(gmm.score_samples(data)[0])

#         if loglk > max_loglk:
#             best_gmm = gmm
#             max_loglk = loglk

#     return best_gmm


# # Inverse Simpson Index, or the effective diversity of power 2
# def effective_diversity(counts):
#     numerator = np.sum(counts)**2
#     denominator = np.sum(v**2 for v in counts)
#     effective_diversity = tk_stats.robust_divide(float(numerator), float(denominator))
#     return effective_diversity


# def compute_percentile_from_distribution(counter, percentile):
#     """ Takes a Counter object (or value:frequency dict) and computes a single percentile.
#     Uses Type 7 interpolation from:
#       Hyndman, R.J.; Fan, Y. (1996). "Sample Quantiles in Statistical Packages".
#     """
#     assert 0 <= percentile <= 100

#     n = np.sum(counter.values())
#     h = (n - 1) * (percentile / 100.0)
#     lower_value = None

#     cum_sum = 0
#     for value, freq in sorted(counter.items()):
#         cum_sum += freq
#         if cum_sum > np.floor(h) and lower_value is None:
#             lower_value = value
#         if cum_sum > np.ceil(h):
#             return lower_value + (h - np.floor(h)) * (value - lower_value)

# # Test for compute_percentile_from_distribution()
# # def test_percentile(x, p):
# #    c = Counter()
# #    for xi in x:
# #        c[xi] += 1
# #    my_res = np.array([compute_percentile_from_distribution(c, p_i) for p_i in p], dtype=float)
# #    numpy_res = np.percentile(x, p)
# #    print np.sum(np.abs(numpy_res - my_res))

# def compute_iqr_from_distribution(counter):
#     p25 = compute_percentile_from_distribution(counter, 25)
#     p75 = compute_percentile_from_distribution(counter, 75)
#     return p75 - p25


# def compute_median_from_distribution(counter):
#     return compute_percentile_from_distribution(counter, 50)


# def correct_bc_error(bc_confidence_threshold, seq, qual, wl_dist):
#     '''Attempt to correct an incorrect BC sequence by computing
#     the probability that a Hamming distance=1 BC generated
#     the observed sequence, accounting for the prior distribution
#     of the whitelist barcodes (wl_dist), and the QV of the base
#     that must have been incorrect'''

#     # QV values
#     qvs = np.fromstring(qual, dtype=np.byte) - tk_constants.ILLUMINA_QUAL_OFFSET

#     # Char array of read
#     a = array.array('c', seq)

#     # Likelihood of candidates
#     wl_cand = []
#     likelihoods = []

#     # Enumerate Hamming distance 1 sequences - if a sequence
#     # is on the whitelist, compute it's likelihood.
#     for pos in range(len(a)):
#         existing = a[pos]
#         for c in tk_seq.NUCS:
#             if c == existing:
#                 continue
#             a[pos] = c
#             test_str = a.tostring()

#             # prior probability of this BC
#             p_bc = wl_dist.get(test_str)
#             if p_bc is not None:
#                 # probability of the base error
#                 edit_qv = min(33.0, float(qvs[pos]))
#                 p_edit = 10.0**(-edit_qv / 10.0)
#                 wl_cand.append(test_str)
#                 likelihoods.append(p_bc * p_edit)

#         a[pos] = existing

#     posterior = np.array(likelihoods)
#     posterior /= posterior.sum()
#     if len(posterior) > 0:
#         pmax = posterior.max()
#         if pmax > bc_confidence_threshold:
#             return wl_cand[np.argmax(posterior)]

#     return None


def determine_max_filtered_bcs(recovered_cells):
    """ Determine the max # of cellular barcodes to consider """
#     return float(recovered_cells) * cr_constants.FILTER_BARCODES_MAX_RECOVERED_CELLS_MULTIPLE # 6
    return float(recovered_cells) * 6


def init_barcode_filter_result():
    return {
        'filtered_bcs': 0,
        'filtered_bcs_lb': 0,
        'filtered_bcs_ub': 0,
        'max_filtered_bcs': 0,
        'filtered_bcs_var': 0,
        'filtered_bcs_cv': 0,
    }


def summarize_bootstrapped_top_n(top_n_boot):
    top_n_bcs_mean = np.mean(top_n_boot)
    top_n_bcs_sd = np.std(top_n_boot)
    top_n_bcs_var = np.var(top_n_boot)
    result = {}
    result['filtered_bcs_var'] = top_n_bcs_var
#     result['filtered_bcs_cv'] = tk_stats.robust_divide(top_n_bcs_sd, top_n_bcs_mean)
    result['filtered_bcs_lb'] = round(sp_stats.norm.ppf(0.025, top_n_bcs_mean, top_n_bcs_sd))
    result['filtered_bcs_ub'] = round(sp_stats.norm.ppf(0.975, top_n_bcs_mean, top_n_bcs_sd))
    result['filtered_bcs'] = int(round(top_n_bcs_mean))
    return result


def find_within_ordmag(x, baseline_idx):
    x_ascending = np.sort(x)
    baseline = x_ascending[-baseline_idx]
    cutoff = max(1, round(0.1 * baseline))
    # Return the index corresponding to the cutoff in descending order
    return len(x) - np.searchsorted(x_ascending, cutoff)


def filter_cellular_barcodes_ordmag(bc_counts, recovered_cells):
    """ Simply take all barcodes that are within an order of magnitude of a top barcode
        that likely represents a cell
    """
    if recovered_cells is None:
        # Modified parameter, didn't use the default value
        recovered_cells = 3000
#         recovered_cells = cr_constants.DEFAULT_RECOVERED_CELLS_PER_GEM_GROUP # 3000

    # Initialize filter result metrics
    metrics = init_barcode_filter_result()
    # determine max # of cellular barcodes to consider
    max_filtered_bcs = determine_max_filtered_bcs(recovered_cells)
    metrics['max_filtered_bcs'] = max_filtered_bcs

    nonzero_bc_counts = bc_counts[bc_counts > 0]
    if len(nonzero_bc_counts) == 0:
        msg = "WARNING: All barcodes do not have enough reads for ordmag, allowing no bcs through"
        return [], metrics, msg

#     baseline_bc_idx = int(round(float(recovered_cells) * (1 - cr_constants.ORDMAG_RECOVERED_CELLS_QUANTILE))) # Quantile=0.99
    baseline_bc_idx = int(round(float(recovered_cells) * (1 - 0.99)))  # Quantile=0.99
    baseline_bc_idx = min(baseline_bc_idx, len(nonzero_bc_counts) - 1)
    assert baseline_bc_idx < max_filtered_bcs

    # Bootstrap sampling; run algo with many random samples of the data
    top_n_boot = np.array([
        find_within_ordmag(np.random.choice(nonzero_bc_counts, len(nonzero_bc_counts)), baseline_bc_idx)
        for i in range(100)  # 100
        #         for i in range(cr_constants.ORDMAG_NUM_BOOTSTRAP_SAMPLES) # 100
    ])

    metrics.update(summarize_bootstrapped_top_n(top_n_boot))

    # Get the filtered barcodes
    top_n = metrics['filtered_bcs']
    top_bc_idx = np.sort(np.argsort(bc_counts)[::-1][:top_n])
    return top_bc_idx, metrics, None


def filter_cellular_barcodes_fixed_cutoff(bc_counts, cutoff):
    nonzero_bcs = len(bc_counts[bc_counts > 0])
    top_n = min(cutoff, nonzero_bcs)
    # np.argsort(bc_counts) => the indices that would sort an array
    # np.argsort(bc_counts)[0] => idx of the smallest element in array
    # np.argsort(bc_counts)[-1] => idx of the largest element in array
    top_bc_idx = np.sort(np.argsort(bc_counts)[::-1][:top_n])
    metrics = {
        'filtered_bcs': top_n,
        'filtered_bcs_lb': top_n,
        'filtered_bcs_ub': top_n,
        'max_filtered_bcs': 0,
        'filtered_bcs_var': 0,
        'filtered_bcs_cv': 0,
    }
    return top_bc_idx, metrics, None


# def filter_cellular_barcodes_manual(matrix, cell_barcodes):
#     """ Take take all barcodes that were given as cell barcodes """
#     barcodes = list(set(matrix.bcs) & set(cell_barcodes))

#     metrics = {
#         'filtered_bcs': len(barcodes),
#         'filtered_bcs_lb': len(barcodes),
#         'filtered_bcs_ub': len(barcodes),
#         'max_filtered_bcs': 0,
#         'filtered_bcs_var': 0,
#         'filtered_bcs_cv': 0,
#     }

#     return barcodes, metrics, None


# def merge_filtered_metrics(filtered_metrics):
#     result = {
#         'filtered_bcs': 0,
#         'filtered_bcs_lb': 0,
#         'filtered_bcs_ub': 0,
#         'max_filtered_bcs': 0,
#         'filtered_bcs_var': 0,
#         'filtered_bcs_cv': 0,
#     }
#     for i, fm in enumerate(filtered_metrics):
#         # Add per-gem group metrics
#         result.update({'gem_group_%d_%s' % (i + 1, key): value for key, value in fm.iteritems()})

#         # Compute metrics over all gem groups
#         result['filtered_bcs'] += fm['filtered_bcs']
#         result['filtered_bcs_lb'] += fm['filtered_bcs_lb']
#         result['filtered_bcs_ub'] += fm['filtered_bcs_ub']
#         result['max_filtered_bcs'] += fm['max_filtered_bcs']
#         result['filtered_bcs_var'] += fm['filtered_bcs_var']

#     # Estimate CV based on sum of variances and means
#     result['filtered_bcs_cv'] = tk_stats.robust_divide(
#         np.sqrt(result['filtered_bcs_var']), fm['filtered_bcs'])

#     return result

# def correct_umis(dupe_keys):
#     corrected_dupe_keys = collections.defaultdict(dict)
#     for dupe_key, umis in dupe_keys.iteritems():
#         for umi in umis:
#             new_umi = correct_umi(umi, umis)
#             if not (new_umi == umi):
#                 corrected_dupe_keys[dupe_key][umi] = new_umi

#     return corrected_dupe_keys

# def correct_umi(seq, counts):
#     corrected_seq = seq
#     count = counts.get(seq, 0)

#     a = array.array('c', seq)
#     for pos in xrange(len(a)):
#         existing = a[pos]
#         for c in tk_seq.NUCS:
#             if c == existing:
#                 continue
#             a[pos] = c
#             test_str = a.tostring()

#             value = counts.get(test_str, 0)
#             if value > count or (value == count and corrected_seq < test_str):
#                 corrected_seq = test_str
#                 count = value

#         a[pos] = existing
#     return corrected_seq


def est_background_profile_bottom(matrix, bottom_frac):
    """Construct a background expression profile from the barcodes that make up the bottom b% of the data
    Args:
      matrix (scipy.sparse.csc_matrix): Feature x Barcode matrix
      bottom_frac (float): Use barcodes making up the bottom x fraction of the counts (0-1)
    Returns:
      (nz_feat (ndarray(int)), profile_p (ndarray(float)): Indices of nonzero features and background profile
    """
    assert bottom_frac >= 0 and bottom_frac <= 1
    umis_per_bc = np.ravel(np.asarray(matrix.sum(0)))
    barcode_order = np.argsort(umis_per_bc)

    cum_frac = np.cumsum(umis_per_bc[barcode_order]) / float(umis_per_bc.sum())
    max_bg_idx = np.searchsorted(cum_frac, bottom_frac, side='left')
    bg_mat = matrix[:, barcode_order[0:max_bg_idx]]

    nz_feat = np.flatnonzero(np.asarray(bg_mat.sum(1)))
    bg_profile = np.ravel(bg_mat[nz_feat, :].sum(axis=1))
    bg_profile_p = bg_profile / float(np.sum(bg_profile))
    assert np.isclose(bg_profile_p.sum(), 1)

    return (nz_feat, bg_profile_p)


def eval_multinomial_loglikelihoods(matrix, profile_p, max_mem_gb=0.1):
    """Compute the multinomial log PMF for many barcodes
    Args:
      matrix (scipy.sparse.csc_matrix): Matrix of UMI counts (feature x barcode)
      profile_p (np.ndarray(float)): Multinomial probability vector
      max_mem_gb (float): Try to bound memory usage.
    Returns:
      log_likelihoods (np.ndarray(float)): Log-likelihood for each barcode
    """
    gb_per_bc = float(matrix.shape[0] * matrix.dtype.itemsize) / (1024**3)
    bcs_per_chunk = max(1, int(round(max_mem_gb/gb_per_bc)))
    num_bcs = matrix.shape[1]

    loglk = np.zeros(num_bcs)

    for chunk_start in range(0, num_bcs, bcs_per_chunk):
        chunk = slice(chunk_start, chunk_start+bcs_per_chunk)
        matrix_chunk = matrix[:, chunk].transpose().toarray()
        n = matrix_chunk.sum(1)
        loglk[chunk] = sp_stats.multinomial.logpmf(matrix_chunk, n, p=profile_p)
    return loglk


def simulate_multinomial_loglikelihoods(profile_p, umis_per_bc,
                                        num_sims=1000, jump=1000,
                                        n_sample_feature_block=1000000, verbose=False):
    """Simulate draws from a multinomial distribution for various values of N.
       Uses the approximation from Lun et al. ( https://www.biorxiv.org/content/biorxiv/early/2018/04/04/234872.full.pdf )
    Args:
      profile_p (np.ndarray(float)): Probability of observing each feature.
      umis_per_bc (np.ndarray(int)): UMI counts per barcode (multinomial N).
      num_sims (int): Number of simulations per distinct N value.
      jump (int): Vectorize the sampling if the gap between two distinct Ns exceeds this.
      n_sample_feature_block (int): Vectorize this many feature samplings at a time.
    Returns:
      (distinct_ns (np.ndarray(int)), log_likelihoods (np.ndarray(float)):
      distinct_ns is an array containing the distinct N values that were simulated.
      log_likelihoods is a len(distinct_ns) x num_sims matrix containing the
        simulated log likelihoods.
    """
    distinct_n = np.flatnonzero(np.bincount(umis_per_bc))

    loglk = np.zeros((len(distinct_n), num_sims), dtype=float)
    num_all_n = np.max(distinct_n) - np.min(distinct_n)
    if verbose:
        print('Number of distinct N supplied: %d' % len(distinct_n))
        print('Range of N: %d' % num_all_n)
        print('Number of features: %d' % len(profile_p))

    sampled_features = np.random.choice(len(profile_p), size=n_sample_feature_block, p=profile_p, replace=True)
    k = 0

    log_profile_p = np.log(profile_p)

    for sim_idx in range(num_sims):
        if verbose and sim_idx % 100 == 99:
            sys.stdout.write('.')
            sys.stdout.flush()
        curr_counts = np.ravel(sp_stats.multinomial.rvs(distinct_n[0], profile_p, size=1))

        curr_loglk = sp_stats.multinomial.logpmf(curr_counts, distinct_n[0], p=profile_p)

        loglk[0, sim_idx] = curr_loglk

        for i in range(1, len(distinct_n)):
            step = distinct_n[i] - distinct_n[i-1]
            if step >= jump:
                # Instead of iterating for each n, sample the intermediate ns all at once
                curr_counts += np.ravel(sp_stats.multinomial.rvs(step, profile_p, size=1))
                curr_loglk = sp_stats.multinomial.logpmf(curr_counts, distinct_n[i], p=profile_p)
                assert not np.isnan(curr_loglk)
            else:
                # Iteratively sample between the two distinct values of n
                for n in range(distinct_n[i-1]+1, distinct_n[i]+1):
                    j = sampled_features[k]
                    k += 1
                    if k >= n_sample_feature_block:
                        # Amortize this operation
                        sampled_features = np.random.choice(
                            len(profile_p), size=n_sample_feature_block, p=profile_p, replace=True)
                        k = 0
                    curr_counts[j] += 1
                    curr_loglk += log_profile_p[j] + np.log(float(n)/curr_counts[j])

            loglk[i, sim_idx] = curr_loglk

    if verbose:
        sys.stdout.write('\n')

    return distinct_n, loglk


def compute_ambient_pvalues(umis_per_bc, obs_loglk, sim_n, sim_loglk):
    """Compute p-values for observed multinomial log-likelihoods
    Args:
      umis_per_bc (nd.array(int)): UMI counts per barcode
      obs_loglk (nd.array(float)): Observed log-likelihoods of each barcode deriving from an ambient profile
      sim_n (nd.array(int)): Multinomial N for simulated log-likelihoods
      sim_loglk (nd.array(float)): Simulated log-likelihoods of shape (len(sim_n), num_simulations)
    Returns:
      pvalues (nd.array(float)): p-values
    """
    assert len(umis_per_bc) == len(obs_loglk)
    assert sim_loglk.shape[0] == len(sim_n)

    # Find the index of the simulated N for each barcode
    sim_n_idx = np.searchsorted(sim_n, umis_per_bc)
    num_sims = sim_loglk.shape[1]

    num_barcodes = len(umis_per_bc)

    pvalues = np.zeros(num_barcodes)

    for i in range(num_barcodes):
        num_lower_loglk = np.sum(sim_loglk[sim_n_idx[i], :] < obs_loglk[i])
        pvalues[i] = float(1 + num_lower_loglk) / (1 + num_sims)
    return pvalues



import numpy as np
import pandas as pd

import celescope.tools.utils as utils
from celescope.__init__ import HELP_DICT
from celescope.tools.step import Step, s_common
from celescope.vdj.__init__ import CHAINS


# UMI_min auto = CELL_CALLING_RANK UMI / 10
CELL_CALLING_RANK = 20
# mixcr sequence header
SEQUENCES_HEADER = ["aaSeqCDR3", "nSeqCDR3"]


class Count_vdj(Step):
    """
    Features
    - Cell-calling based on barcode-UMI rank.    
    - Summarize clonetypes infomation.

    Output
    - `{sample}_cell_confident.tsv` The clone type of VDJ cell barcode, each chain occupies one line.

    - `{sample}_cell_confident_count.tsv` The clone type of VDJ cell barcode, each cell occupies one line.

    - `{sample}_clonetypes.tsv` The count and percentage of each clonetypes of VDJ cell barcode.

    - `{sample}_match_clonetypes.tsv` When summarize clonetypes, only consider barcodes in the match scRNA-Seq library. 
    This file will only be produced when the `match_dir` parameter is provided.
    """

    def __init__(self, args, step_name):
        Step.__init__(self, args, step_name)

        # set
        self.chains = CHAINS[args.type]
        self.cols = []
        for chain in self.chains:
            for seq in SEQUENCES_HEADER:
                self.cols.append("_".join([seq, chain]))

        self.match_bool = False
        if args.match_dir and args.match_dir.strip() != 'None':
            self.match_cell_barcodes, _match_cell_number = utils.read_barcode_file(
                args.match_dir)
            self.match_bool = True
        elif args.matrix_dir and args.matrix_dir.strip() != 'None':
            self.match_cell_barcodes= utils.get_barcodes_from_matrix_dir(args.matrix_dir)            
            self.match_bool = True
        if self.match_bool:
            self.match_cell_barcodes = set(self.match_cell_barcodes)

        # out files
        self.cell_confident_file = f"{self.out_prefix}_cell_confident.tsv"
        self.cell_confident_count_file = f"{self.out_prefix}_cell_confident_count.tsv"
        self.clonetypes_file = f"{self.out_prefix}_clonetypes.tsv"
        self.match_clonetypes_file = f"{self.out_prefix}_match_clonetypes.tsv"

        # add args data
        self.add_data_item(iUMI=args.iUMI)

    @utils.add_log
    def cell_calling(self, df_UMI_count_filter):
        df_UMI_sum = df_UMI_count_filter.groupby(
            ['barcode'], as_index=False).agg({"UMI": "sum"})
        if (self.args.UMI_min == "auto"):
            df_UMI_sum_sorted = df_UMI_sum.sort_values(
                ["UMI"], ascending=False)
            rank_UMI = df_UMI_sum_sorted.iloc[CELL_CALLING_RANK, :]["UMI"]
            UMI_min = int(rank_UMI / 10)
        else:
            UMI_min = int(self.args.UMI_min)
        df_UMI_cell = df_UMI_sum[df_UMI_sum.UMI >= UMI_min]
        df_UMI_sum["mark"] = df_UMI_sum["UMI"].apply(
            lambda x: "CB" if (x >= UMI_min) else "UB")

        df = df_UMI_sum.sort_values('UMI', ascending=False)
        self.add_data_item(CB_num=df[df['mark'] == 'CB'].shape[0])
        self.add_data_item(Cells=list(df.loc[df['mark'] == 'CB', 'UMI']))
        self.add_data_item(UB_num=df[df['mark'] == 'UB'].shape[0])
        self.add_data_item(Background=list(df.loc[df['mark'] == 'UB', 'UMI']))

        cell_barcodes = set(df_UMI_cell.barcode)
        total_cell_number = len(cell_barcodes)
        self.add_metric(
            name="Estimated Number of Cells",
            value=total_cell_number,
        )

        df_cell = df_UMI_count_filter[df_UMI_count_filter.barcode.isin(
            cell_barcodes)]
        return df_cell, cell_barcodes

    @utils.add_log
    def get_df_confident(self, df_cell):
        """
        1. UMI > iUMI
        2. in chain
        """
        df_iUMI = df_cell[df_cell.UMI >= self.args.iUMI]
        df_confident = df_iUMI[df_iUMI["chain"].isin(self.chains)]
        df_confident = df_confident.sort_values("UMI", ascending=False)
        df_confident = df_confident.groupby(
            ["barcode", "chain"], as_index=False).head(1)
        return df_confident

    def get_df_valid_count(self, df_confident):
        df_valid_count = df_confident.set_index(["barcode", "chain"])
        df_valid_count = df_valid_count.unstack()
        df_valid_count.columns = ['_'.join(col) for col in df_valid_count]
        df_valid_count = df_valid_count.reset_index()
        df_valid_count.fillna(inplace=True, value="NA")
        return df_valid_count

    def get_clonetypes_and_write(self, df_valid_count, cell_barcodes):
        """
        Returns
        - df_clonetypes
        - df_match_clonetypes
        """

        total_cell_number = len(cell_barcodes)
        df_clonetypes = df_valid_count.copy()
        df_match_clonetypes = None

        df_clonetypes = df_clonetypes.groupby(self.cols, as_index=False).agg({
            "barcode": "count"})
        # put na last
        df_clonetypes.replace('NA', np.nan, inplace=True)
        df_clonetypes.sort_values(
            ["barcode"] + self.cols, ascending=False, na_position='last', inplace=True)
        df_clonetypes.replace(np.nan, 'NA', inplace=True)

        total_CDR3_barcode_number = sum(df_clonetypes.barcode)
        df_clonetypes["percent"] = df_clonetypes.barcode / \
            total_CDR3_barcode_number * 100
        df_clonetypes["percent"] = df_clonetypes["percent"].apply(
            lambda x: round(x, 2))

        # add clonetype ID
        df_clonetypes = df_clonetypes.reset_index()
        df_clonetypes["clonetype_ID"] = pd.Series(df_clonetypes.index) + 1
        df_clonetypes.drop(columns=["index"], inplace=True)

        # order
        order = ["clonetype_ID"] + self.cols + ["barcode", "percent"]
        df_clonetypes = df_clonetypes[order]
        df_clonetypes.rename(
            columns={"barcode": "barcode_count"}, inplace=True)
        # out clonetypes
        df_clonetypes.to_csv(self.clonetypes_file, sep="\t", index=False)

        if self.args.type == "TCR":

            UMI_col_dic = {"TRA": "UMI_TRA", "TRB": "UMI_TRB"}
            for chain in UMI_col_dic:
                UMI_col_name = UMI_col_dic[chain]
                if UMI_col_name in df_valid_count.columns:
                    df_valid_count[UMI_col_name].replace(
                        "NA", 0, inplace=True)
                    Median_chain_UMIs_per_Cell = np.median(
                        df_valid_count[UMI_col_name])
                else:
                    Median_chain_UMIs_per_Cell = 0
                self.add_metric(
                    name=f"Median {chain} UMIs per Cell",
                    value=Median_chain_UMIs_per_Cell,
                )

            df_TRA_TRB = df_valid_count[
                (df_valid_count.aaSeqCDR3_TRA != "NA") &
                (df_valid_count.aaSeqCDR3_TRB != "NA")
            ]
            cell_with_confident_TRA_and_TRB = df_TRA_TRB.shape[0]
            self.add_metric(
                name="Cell with TRA and TRB",
                value=cell_with_confident_TRA_and_TRB,
                total=total_cell_number
            )

            if self.match_bool:
                cell_with_match_barcode = self.match_cell_barcodes.intersection(
                    cell_barcodes)
                cell_with_match_barcode_number = len(cell_with_match_barcode)

                df_match = df_valid_count[df_valid_count.barcode.isin(
                    self.match_cell_barcodes)]

                df_match_TRA_TRB = df_match[
                    (df_match.aaSeqCDR3_TRA != "NA") &
                    (df_match.aaSeqCDR3_TRB != "NA")
                ]
                match_cell_with_TRA_and_TRB = df_match_TRA_TRB.shape[0]
                self.add_metric(
                    name="Cell with Barcode Match",
                    value=cell_with_match_barcode_number,
                    total=total_cell_number
                )
                self.add_metric(
                    name="Cell with Barcode Match, TRA and TRB",
                    value=match_cell_with_TRA_and_TRB,
                    total=total_cell_number
                )

        # BCR
        elif self.args.type == "BCR":

            UMI_col_dic = {"IGH": "UMI_IGH",
                           "IGL": "UMI_IGL", "IGK": "UMI_IGK"}
            for chain in UMI_col_dic:
                UMI_col_name = UMI_col_dic[chain]
                if UMI_col_name in df_valid_count.columns:
                    df_valid_count[UMI_col_name].replace(
                        "NA", 0, inplace=True)
                    df_valid_count_over_zero = df_valid_count[
                        df_valid_count[UMI_col_name] > 0
                    ]
                    Median_chain_UMIs_per_Cell = np.median(
                        df_valid_count_over_zero[UMI_col_name])
                else:
                    Median_chain_UMIs_per_Cell = 0
                self.add_metric(
                    name=f"Median {chain} UMIs per Cell",
                    value=Median_chain_UMIs_per_Cell,
                )

            df_heavy_and_light = df_valid_count[
                (df_valid_count.aaSeqCDR3_IGH != "NA") &
                (
                    (df_valid_count.aaSeqCDR3_IGL != "NA") |
                    (df_valid_count.aaSeqCDR3_IGK != "NA")
                )
            ]
            Cell_with_Heavy_and_Light_Chain = df_heavy_and_light.shape[0]
            self.add_metric(
                name="Cell with Heavy and Light Chain",
                value=Cell_with_Heavy_and_Light_Chain,
                total=total_cell_number
            )

            if self.match_bool:
                cell_with_match_barcode = self.match_cell_barcodes.intersection(
                    cell_barcodes)
                cell_with_match_barcode_number = len(cell_with_match_barcode)

                df_match = df_valid_count[df_valid_count.barcode.isin(
                    self.match_cell_barcodes)]

                # median match UMI
                df_match_heavy_light = df_match[
                    (df_match.aaSeqCDR3_IGH != "NA") &
                    (
                        (df_match.aaSeqCDR3_IGL != "NA") |
                        (df_match.aaSeqCDR3_IGK != "NA")
                    )
                ]
                match_cell_with_heavy_and_light = df_match_heavy_light.shape[0]
                self.add_metric(
                    name="Cell with Barcode Match",
                    value=cell_with_match_barcode_number,
                    total=total_cell_number
                )
                self.add_metric(
                    name="Cell with Barcode Match, Heavy and Light Chain",
                    value=match_cell_with_heavy_and_light,
                    total=total_cell_number
                )

        if self.match_bool:
            """
            df_match_clonetypes
            """
            df_match_clonetypes = df_match.groupby(self.cols, as_index=False).agg({
                "barcode": "count"})
            total_match_CDR3_barcode_number = sum(
                df_match_clonetypes.barcode)
            df_match_clonetypes["percent"] = df_match_clonetypes.barcode / \
                total_match_CDR3_barcode_number * 100
            df_match_clonetypes["percent"] = df_match_clonetypes["percent"].apply(
                lambda x: round(x, 2)
            )
            df_match_clonetypes.rename(
                columns={"barcode": "barcode_count"}, inplace=True)
            df_match_clonetypes = df_match_clonetypes.merge(
                df_clonetypes, on=self.cols, how='left', suffixes=('', '_y'))
            # order and drop duplicated cols
            order = ["clonetype_ID"] + self.cols + ["barcode_count", "percent"]
            df_match_clonetypes = df_match_clonetypes[order]
            df_match_clonetypes.sort_values(["barcode_count", "clonetype_ID"], ascending=[
                                            False, True], inplace=True)
            df_match_clonetypes.to_csv(
                self.match_clonetypes_file, sep="\t", index=False)
        return df_clonetypes, df_match_clonetypes

    def write_cell_confident_count(self, df_valid_count, df_clonetypes, df_confident):
        df_mergeID = pd.merge(df_valid_count,
                              df_clonetypes, how="left", on=self.cols)
        df_mergeID.sort_values(["clonetype_ID", "barcode"], inplace=True)
        # output df_valid_count
        df_mergeID.to_csv(self.cell_confident_count_file,
                          sep="\t", index=False)
        df_mergeID = df_mergeID[["barcode", "clonetype_ID"]]
        df_cell_confident_with_ID = pd.merge(
            df_confident, df_mergeID, how="left", on="barcode")
        df_cell_confident_with_ID.sort_values(
            ["clonetype_ID", "barcode", "chain"], inplace=True)
        # output df_cell_confident
        df_cell_confident_with_ID.to_csv(
            self.cell_confident_file, sep="\t", index=False)

    def write_clonetypes_table_to_data(self, df_clonetypes, df_match_clonetypes):
        # cloneytpes table
        def format_table(df_clonetypes):
            df_table = df_clonetypes.copy()
            df_table["percent"] = df_table["percent"].apply(
                lambda x: str(x) + "%")
            seqs = ["aaSeqCDR3"]
            cols = []
            for chain in self.chains:
                for seq in seqs:
                    cols.append("_".join([seq, chain]))
            df_table_cols = ["clonetype_ID"] + \
                cols + ["barcode_count", "percent"]
            df_table = df_table[df_table_cols]
            table_header = ["Clonetype_ID"] + cols + ["Frequency", "Percent"]
            return df_table, table_header

        df_table, _table_header = format_table(df_clonetypes)
        title = 'Clonetypes'
        if self.match_bool:
            df_table, _table_header = format_table(df_match_clonetypes)
            title = 'Match Clonetypes'

        table_dict = self.get_table(title, 'clonetypes_table', df_table)
        self.add_data_item(table_dict=table_dict)

    def run(self):
        df_UMI_count_filter = pd.read_csv(
            self.args.UMI_count_filter_file, sep='\t')
        df_cell, cell_barcodes = self.cell_calling(df_UMI_count_filter)
        df_confident = self.get_df_confident(df_cell)
        df_valid_count = self.get_df_valid_count(df_confident)
        df_clonetypes, df_match_clonetypes = self.get_clonetypes_and_write(
            df_valid_count, cell_barcodes)
        self.write_cell_confident_count(
            df_valid_count, df_clonetypes, df_confident)
        self.write_clonetypes_table_to_data(df_clonetypes, df_match_clonetypes)
        self.clean_up()


def count_vdj(args):
    # TODO
    # add TCR or BCR prefix to distinguish them in html report summary; should improve
    step_name = f"{args.type}_count_vdj"
    count_vdj_obj = Count_vdj(args, step_name)
    count_vdj_obj.run()


def get_opts_count_vdj(parser, sub_program):
    parser.add_argument(
        "--type", help="Required. `TCR` or `BCR`. ", required=True)
    parser.add_argument(
        '--UMI_min',
        help='Default `auto`. Minimum UMI number to filter. The barcode with UMI>=UMI_min is considered to be cell.',
        default="auto"
    )
    parser.add_argument(
        '--iUMI',
        help="""Default `1`. Minimum number of UMI of identical receptor type and CDR3. 
For each (barcode, chain) combination, only UMI>=iUMI is considered valid.""",
        type=int,
        default=1
    )
    if sub_program:
        parser.add_argument("--UMI_count_filter_file",
                            help="Required. File from step mapping_vdj.", required=True)
        parser.add_argument("--match_dir", help=HELP_DICT['match_dir'])
        parser.add_argument("--matrix_dir", help=HELP_DICT['matrix_dir'])
        parser = s_common(parser)



import os

import editdistance
import pandas as pd
import pysam

from celescope.tools.utils import add_log, parse_match_dir

parentDir = os.path.dirname(__file__)


def read_mut(mut_file):
    mut_dic = {}
    df = pd.read_csv(mut_file, sep="\t")

    def convert_row(row):
        gene = row["gene"]
        mut_type = row["type"]
        seq = row["seq"]
        ref_position = row["ref_position"]
        if gene not in mut_dic:
            mut_dic[gene] = {}
        mut_dic[gene][mut_type] = {}
        mut_dic[gene][mut_type]["pos"] = ref_position
        mut_dic[gene][mut_type]["seq"] = seq
    df.apply(func=convert_row, axis=1)
    return mut_dic


def is_fusion(pos, read_start, read_length, flanking_base):
    test_start = (pos - flanking_base) >= read_start
    test_end = (pos + flanking_base) <= (read_start + read_length)
    return (test_start and test_end)


@add_log
def count_mut(args):

    outdir = args.outdir
    sample = args.sample
    bam = args.bam
    shift_base = int(args.shift_base)
    mut_file = args.mut_file
    match_dir = args.match_dir

    # check dir
    if not os.path.exists(outdir):
        os.system('mkdir -p %s' % (outdir))

    mut_dic = read_mut(mut_file)
    out_prefix = outdir + "/" + sample

    # tsne
    match_dict = parse_match_dir(match_dir)
    df_tsne = pd.read_csv(match_dict['tsne_coord'], sep="\t", index_col=0)

    # out
    out_read_file = out_prefix + "_mut_read.tsv"
    out_read_count_file = out_prefix + "_mut_read_count.tsv"
    out_umi_count_file = out_prefix + "_mut_UMI_count.tsv"
    out_insertion_barcode_count_file = out_prefix + "_mut_insertion_barcode_count.tsv"
    out_tsne_file = out_prefix + "_mut_tsne.tsv"

    # process bam
    samfile = pysam.AlignmentFile(bam, "rb")
    # header = samfile.header
    # new_bam = pysam.AlignmentFile(out_prefix+"_mut.bam", "wb", header=header)
    rows = []
    for read in samfile:
        tag = read.reference_name
        attr = read.query_name.split('_')
        barcode = attr[0]
        umi = attr[1]
        seq = read.query_sequence
        cigar = read.cigar
        ref_pos = read.reference_start
        read_pos = read.query_alignment_start
        for (cigarType, cigarLength) in cigar:
            if cigarType == 0:  # match
                ref_pos += cigarLength
                read_pos += cigarLength
            elif cigarType == 1:  # insertion
                insert_seq = seq[read_pos:read_pos+cigarLength]
                insert_seq_length = len(insert_seq)
                rows.append({
                    "barcode": barcode,
                    "UMI": umi,
                    "gene": tag,
                    "type": "insertion",
                    "seq": insert_seq,
                    "ref_pos": ref_pos,
                    "read_pos": read_pos,
                    "seq_length": insert_seq_length
                })
                read_pos += cigarLength
            elif cigarType == 2:  # deletion
                ref_pos += cigarLength
    df = pd.DataFrame(rows)
    df = df[["barcode", "UMI", "gene", "type",
             "ref_pos", "read_pos", "seq", "seq_length"]]

    def is_valid(row):
        gene = str(row["gene"])
        ref_pos = int(row["ref_pos"])
        mut_type = str(row["type"])
        int(row["seq_length"])
        seq = str(row["seq"])
        if gene not in mut_dic:
            return False
        if mut_type not in mut_dic[gene].keys():
            return False
        if abs(ref_pos-mut_dic[gene][mut_type]["pos"]) > shift_base:
            return False
        if editdistance.eval(seq, mut_dic[gene][mut_type]["seq"]) > shift_base:
            return False
        return True

    df["is_valid"] = df.apply(func=is_valid, axis=1)
    df.to_csv(out_read_file, sep="\t")

    df_valid = df[df["is_valid"]]
    df_read_count = df_valid.groupby(
        ["gene", "type", "barcode", "UMI"]).agg({"UMI": "count"})
    df_read_count.columns = ["read_count"]
    df_read_count.reset_index(inplace=True)
    df_read_count.to_csv(out_read_count_file, sep="\t")

    df_UMI_count = df_read_count.groupby(
        ["gene", "type", "barcode"]).agg({"UMI": "count"})
    df_UMI_count.columns = ["UMI_count"]
    df_UMI_count.to_csv(out_umi_count_file, sep="\t")

    df_temp = df_UMI_count.reset_index()
    if df_temp.shape[0] == 0:
        count_mut.logger.warning('NO VALID INSERTION FOUND!')
    else:
        df_insertion = df_temp[df_temp["type"] == "insertion"]
        df_insertion_barcode_count = df_insertion.pivot(
            index="barcode", columns="gene", values="UMI_count")
        df_insertion_barcode_count.to_csv(
            out_insertion_barcode_count_file, sep="\t")

        df_tsne_mut = pd.merge(df_tsne, df_insertion_barcode_count,
                               right_index=True, left_index=True, how="left")
        df_tsne_mut.fillna(0, inplace=True)
        df_tsne_mut.to_csv(out_tsne_file, sep="\t")

        # plot
        app = parentDir + "/plot.R"
        cmd = f"Rscript {app} --tsne_mut {out_tsne_file} --outdir {outdir}"
        count_mut.logger.info(cmd)
        os.system(cmd)
        count_mut.logger.info("plot done.")


def get_opts_count_mut(parser, sub_program):
    if sub_program:
        parser.add_argument('--outdir', help='output dir', required=True)
        parser.add_argument('--sample', help='sample name', required=True)
        parser.add_argument("--bam", required=True)
        parser.add_argument('--assay', help='assay', required=True)
    parser.add_argument("--mut_file", help="mutation file", required=True)
    parser.add_argument(
        "--match_dir", help="match scRNA-Seq dir", required=True)
    parser.add_argument("--shift_base", default=2)



"""
map read2 to barcode_fasta
"""

import pandas as pd
import pysam

import celescope.tools.utils as utils
from celescope.tools.barcode import Barcode
from celescope.tools.step import Step, s_common


def get_opts_mapping_tag(parser, sub_program):
    parser.add_argument(
        "--fq_pattern",
        help="""Required. R2 read pattern. The number after the letter represents the number of bases.         
`L` linker(common sequences)  
`C` tag barcode  
""",
        required=True
    )
    parser.add_argument(
        "--barcode_fasta",
        help="""Required. Tag barcode fasta file. It will check the mismatches between tag barcode 
sequence in R2 reads with all tag barcode sequence in barcode_fasta. 
It will assign read to the tag with mismatch < len(tag barcode) / 10 + 1. 
If no such tag exists, the read is classified as invalid.
```
>tag_0
GGGCGTCTGTGACCGCGTGATACTGCATTGTAGACCGCCCAACTC
>tag_1
TTCCTCCAGAGGAGACCGAGCCGGTCAATTCAGGAGAACGTCCGG
>tag_2
AGGGCTAGGCGTGTCATTTGGCGAGGTCCTGAGGTCATGGAGCCA
>tag_3
CACTGGTCATCGACACTGGGAACCTGAGGTGAGTTCGCGCGCAAG
```  
""",
        required=True,
    )
    parser.add_argument(
        "--linker_fasta",
        help="""Optional. If provided, it will check the mismatches between linker sequence in R2 reads 
with all linker sequence in linker_fasta. If no mismatch < len(linker) / 10 + 1, the read is classified as invalid.
""",
    )
    if sub_program:
        s_common(parser)
        parser.add_argument("--fq", help="R2 read fastq.", required=True)


@utils.add_log
def mapping_tag(args):
    step_name = "mapping_tag"
    runner = Mapping_tag(args, step_name)
    runner.run()


class Mapping_tag(Step):
    """
    Features
    - Align R2 reads to the tag barcode fasta.

    Output

    - `{sample}_read_count.tsv` tab-delimited text file with 4 columns.

        `barcode` cell barcode  
        `tag_name`  tag name in barcode_fasta  
        `UMI`   UMI sequence  
        `read_count` read count per UMI  
    """

    def __init__(self, args, step_name):
        Step.__init__(self, args, step_name)

        # read args
        self.fq = args.fq
        self.fq_pattern = args.fq_pattern
        self.linker_fasta = args.linker_fasta
        self.barcode_fasta = args.barcode_fasta

        # process
        self.barcode_dict, self.barcode_length = utils.read_fasta(self.barcode_fasta, equal=True)
        if self.linker_fasta and self.linker_fasta != 'None':
            self.linker_dict, self.linker_length = utils.read_fasta(self.linker_fasta, equal=True)
        else:
            self.linker_dict, self.linker_length = {}, 0
        self.pattern_dict = Barcode.parse_pattern(self.fq_pattern)

        # check barcode length
        barcode1 = self.pattern_dict["C"][0]
        # end - start
        pattern_barcode_length = barcode1[1] - barcode1[0]
        if pattern_barcode_length != self.barcode_length:
            raise Exception(
                f'''barcode fasta length {self.barcode_length} 
                != pattern barcode length {pattern_barcode_length}'''
            )

        self.res_dic = utils.genDict()
        self.res_sum_dic = utils.genDict(dim=2)
        self.match_barcode = []

        # out files
        self.read_count_file = f'{self.outdir}/{self.sample}_read_count.tsv'
        self.UMI_count_file = f'{self.outdir}/{self.sample}_UMI_count.tsv'
        self.stat_file = f'{self.outdir}/stat.txt'

    def process_read(self):
        total_reads = 0
        reads_unmapped_too_short = 0
        reads_unmapped_invalid_iinker = 0
        reads_unmapped_invalid_barcode = 0
        reads_mapped = 0

        with pysam.FastxFile(self.fq) as infile:
            for record in infile:
                total_reads += 1
                attr = str(record.name).strip("@").split("_")
                barcode = str(attr[0])
                umi = str(attr[1])
                seq = record.sequence

                if self.linker_length != 0:
                    seq_linker = utils.seq_ranges(seq, self.pattern_dict['L'])
                    if len(seq_linker) < self.linker_length:
                        reads_unmapped_too_short += 1
                        continue
                if self.barcode_dict:
                    seq_barcode = utils.seq_ranges(seq, self.pattern_dict['C'])
                    if self.barcode_length != len(seq_barcode):
                        miss_length = self.barcode_length - len(seq_barcode)
                        if miss_length > 2:
                            reads_unmapped_too_short += 1
                            continue
                        seq_barcode = seq_barcode + "A" * miss_length

                # check linker
                if self.linker_length != 0:
                    valid_linker = False
                    for linker_name in self.linker_dict:
                        if utils.hamming_correct(self.linker_dict[linker_name], seq_linker):
                            valid_linker = True
                            break
                else:
                    valid_linker = True

                if not valid_linker:
                    reads_unmapped_invalid_iinker += 1
                    continue

                # check barcode
                valid_barcode = False
                for barcode_name in self.barcode_dict:
                    if utils.hamming_correct(self.barcode_dict[barcode_name], seq_barcode):
                        self.res_dic[barcode][barcode_name][umi] += 1
                        valid_barcode = True
                        break

                if not valid_barcode:
                    reads_unmapped_invalid_barcode += 1
                    continue

                # mapped
                reads_mapped += 1

        # write dic to pandas df
        rows = []
        for barcode in self.res_dic:
            for tag_name in self.res_dic[barcode]:
                for umi in self.res_dic[barcode][tag_name]:
                    rows.append([barcode, tag_name, umi,
                                 self.res_dic[barcode][tag_name][umi]])
        df_read_count = pd.DataFrame(rows)
        df_read_count.rename(
            columns={
                0: "barcode",
                1: "tag_name",
                2: "UMI",
                3: "read_count"
            }, inplace=True)
        df_read_count.to_csv(
            self.read_count_file, sep="\t", index=False)

        # add metrics
        self.add_metric(
            name='Reads Mapped',
            value=reads_mapped,
            total=total_reads,
        )
        self.add_metric(
            name='Reads Unmapped too Short',
            value=reads_unmapped_too_short,
            total=total_reads,
        )
        self.add_metric(
            name='Reads Unmapped Invalid Linker',
            value=reads_unmapped_invalid_iinker,
            total=total_reads,
        )
        self.add_metric(
            name='Reads Unmapped Invalid Barcode',
            value=reads_unmapped_invalid_barcode,
            total=total_reads,
        )

    @utils.add_log
    def run(self):
        self.process_read()
        self.clean_up()



import os
from collections import defaultdict

import pandas as pd
import pysam

from celescope.tcr_fl.barcode_index import Barcode_index
from celescope.tools.utils import add_log, genDict, read_barcode_file


@add_log
def get_nCell_barcodes(fq, nCell):
    '''
    get top nCell's barcodes(rank by UMI counts)
    '''
    count_dict = genDict(dim=2)
    barcode_dict = {}
    with pysam.FastxFile(fq) as fh:
        for entry in fh:
            attr = entry.name.split('_')
            barcode = attr[0]
            umi = attr[1]
            count_dict[barcode][umi] += 1
    for barcode in count_dict:
        barcode_dict[barcode] = len(count_dict[barcode])
    barcodes = pd.DataFrame.from_dict(barcode_dict, orient='index').sort_values(
        0, ascending=False).iloc[0:nCell, ].index
    return barcodes


@add_log
def split_run(fq, fq_outdir, barcodes=None, nCell=None):
    '''
    split fastq 
    '''
    if not os.path.exists(fq_outdir):
        os.makedirs(fq_outdir)
    if nCell and nCell != 'None':
        barcodes = get_nCell_barcodes(fq, nCell)
    bi = Barcode_index(barcodes)
    entry_dict = defaultdict(list)
    with pysam.FastxFile(fq) as fh:
        for entry in fh:
            attr = entry.name.split('_')
            barcode = attr[0]
            if barcode in barcodes:
                cell_index = bi.index_dict[barcode]
                entry_dict[cell_index].append(entry)

    # write to file
    for cell_index in entry_dict:
        with open(f'{fq_outdir}/{cell_index}.fq', 'w') as f:
            for entry in entry_dict[cell_index]:
                f.write(str(entry) + '\n')
    return bi


@add_log
def split_fq(args):
    nCell = args.nCell
    outdir = args.outdir
    sample = args.sample
    match_dir = args.match_dir

    if match_dir and match_dir != 'None':
        barcodes, _nCell = read_barcode_file(args.match_dir)
    else:
        barcodes = ''
    fq_outdir = f'{args.outdir}/fastq'
    if nCell and nCell != 'None':
        nCell = int(nCell)
    bi = split_run(args.fq, fq_outdir, barcodes, nCell)
    index_file = f'{outdir}/{sample}_index.tsv'
    bi.df_index.to_csv(index_file, sep='\t')


def get_opts_split_fq(parser, sub_program):
    if sub_program:
        parser.add_argument('--outdir', help='output dir', required=True)
        parser.add_argument('--sample', help='sample name', required=True)
        parser.add_argument("--fq", required=True)
        parser.add_argument('--assay', help='assay', required=True)
    parser.add_argument(
        "--match_dir", help="match scRNA-Seq dir")
    parser.add_argument("--nCell", help="select top N cell")



import pandas as pd
import pysam
import numpy as np

import celescope.tools.utils as utils
from celescope.tools.step import Step,s_common
from celescope.capture_virus.otsu import array2hist, makePlot, threshold_otsu

class Count_capture_virus(Step):
    def __init__(self, args, step_name):
        Step.__init__(self, args, step_name)

        # set
        self.min_query_length = int(args.min_query_length)

        # read barcodes
        self.match_cell_barcodes, _match_cell_number = utils.read_barcode_file(args.match_dir)

        # out
        self.out_read_count_file = f'{self.out_prefix}_virus_read_count.tsv'
        self.out_umi_count_file = f'{self.out_prefix}_virus_UMI_count.tsv'
        self.otsu_plot = f'{self.out_prefix}_otsu.png'


    def otsu_min_support_read(self, array):

        array = np.log10(array)
        hist = array2hist(array)
        thresh = threshold_otsu(hist)
        makePlot(hist, thresh, self.otsu_plot)
        threshold = int(10 ** thresh)
        self.add_content_item(
            slot='metric',
            otsu_min_support_read=threshold,
        )
        return threshold

    @utils.add_log
    def sum_virus(self):
        # process bam
        samfile = pysam.AlignmentFile(self.args.virus_bam, "rb")
        count_dic = utils.genDict(dim=3)
        for read in samfile:
            tag = read.reference_name
            query_length = read.infer_query_length()
            attr = read.query_name.split('_')
            barcode = attr[0]
            umi = attr[1]
            if (barcode in self.match_cell_barcodes) and (query_length >= self.min_query_length):
                count_dic[barcode][tag][umi] += 1

        # write dic to pandas df
        rows = []
        array = []
        for barcode in count_dic:
            for tag in count_dic[barcode]:
                for umi in count_dic[barcode][tag]:
                    read_count = count_dic[barcode][tag][umi]
                    rows.append([barcode, tag, umi, read_count])
                    array.append(read_count)

        df_read = df_read = pd.DataFrame(
            rows,
            columns=[
                "barcode",
                "tag",
                "UMI",
                "read_count"])
        df_read.to_csv(self.out_read_count_file, sep="\t", index=False)

        if self.args.min_support_read == 'auto':
            min_support_read = self.otsu_min_support_read(array)
        else:
            min_support_read = int(self.args.min_support_read)

        df_valid = df_read[df_read["read_count"] >= min_support_read].groupby(
            ["barcode", "tag"]).agg({"UMI": "count"})
        if df_valid.shape[0] == 0:
            self.sum_virus.logger.warning("No cell virus UMI found!")

        df_valid.to_csv(self.out_umi_count_file, sep="\t")

    def run(self):
        self.sum_virus()
        self.dump_content(slot="metric")


@utils.add_log
def count_capture_virus(args):

    step_name = 'count_capture_virus'
    runner = Count_capture_virus(args, step_name)
    runner.run()


def get_opts_count_capture_virus(parser, sub_program):
    parser.add_argument("--min_query_length", help='Minimum query length.', default=35)
    parser.add_argument("--min_support_read", help='Minimum number of reads supporting a UMI', default='auto')
    if sub_program:
        parser.add_argument('--match_dir', help='matched rna_virus directory', required=True)
        parser.add_argument('--virus_bam', required=True)
        s_common(parser)



import glob
import re
import sys
import unittest
from collections import Counter, defaultdict
from itertools import combinations, product

import pysam
from xopen import xopen

from celescope.tools import utils
from celescope.tools.__init__ import PATTERN_DICT
from celescope.__init__ import ROOT_PATH, HELP_DICT
from celescope.tools.step import Step, s_common

MIN_T = 10


class Chemistry():
    """
    Auto detect chemistry from R1-read
    """

    def __init__(self, fq1):
        '''
        'scopeV2.0.1': 'C8L16C8L16C8L1U8T18'
        'scopeV2.1.1': 'C8L16C8L16C8L1U12T18'
        'scopeV2.2.1': 'C8L16C8L16C8L1U12T18' with 4 types of linkers
        'scopeV3.0.1': 'C9L16C9L16C9L1U12T18' with 4 types of linkers
        '''
        self.fq1 = fq1
        self.fq1_list = fq1.split(',')
        self.n_read = 10000

        self.pattern_dict_v2, * \
            _, self.linker_1_v2_set_list, self.linker_1_v2_mismatch_list = Barcode.parse_chemistry('scopeV2.1.1')
        self.pattern_dict_v2, * \
            _, self.linker_4_v2_set_list, self.linker_4_v2_mismatch_list = Barcode.parse_chemistry('scopeV2.2.1')
        self.pattern_dict_v3, *_, self.linker_v3_set_list, self.linker_v3_mismatch_list = Barcode.parse_chemistry('scopeV3.0.1')
        self.pattern_dict_flv, *_, self.linker_flv_set_list, self.linker_flv_mismatch_list = Barcode.parse_chemistry('flv')
        self.pattern_dict_flv_rna, *_, self.linker_flv_rna_set_list, self.linker_flv_rna_mismatch_list = Barcode.parse_chemistry('flv_rna')


    @utils.add_log
    def check_chemistry(self):
        """check chemistry in the fq1_list"""
        chemistry_list = []
        for fastq1 in self.fq1_list:
            print(fastq1)
            chemistry = self.get_chemistry(fastq1)
            chemistry_list.append(chemistry)
        if len(set(chemistry_list)) != 1:
            Chemistry.check_chemistry.logger.warning('multiple chemistry found!' + str(chemistry_list))
        return chemistry_list


    def seq_chemistry(self, seq):
        """
        Returns：chemistry or None
        >>> runner = Chemistry("fake_fq1_string")
        >>> seq = "TCGACTGTCATCCACGTGCTTGAGATTCTAGGATTCAGCATGCGGCTACGTGCACGAGACATATCAATGGGTTTTCTTGTTGCTTTTTTTTTTTTTTTTTTTTTTTT"
        >>> runner.seq_chemistry(seq)
        'scopeV3.0.1'
        >>> seq = "GTCGTAGAATCCACGTGCTTGAGACTCAATGATCAGCATGCGGCTACGGCGATTAACGTTGAATGTTTTTTTTTTTTTTTTTTTTT"
        >>> runner.seq_chemistry(seq)
        'scopeV2.0.1'
        >>> seq = "NCAGATTC" + "ATCCACGTGCTTGAGA" + "GTACGCAA" + "TCAGCATGCGGCTACG" + "CTGAGCCA" + "C" + "TCCGAAGCCCAT" + "TTTTTTTTTTTTTTTTTTTTTTTTTTATTGC"
        >>> runner.seq_chemistry(seq)
        'scopeV2.1.1'
        >>> seq = "NCAGATTC" + "TCGGTGACAGCCATAT" + "GTACGCAA" + "CGTAGTCAGAAGCTGA" + "CTGAGCCA" + "C" + "TCCGAAGCCCAT" + "TTTTTTTTTTTTTTTTTTTTTTTTTTATTGC"
        >>> runner.seq_chemistry(seq)
        'scopeV2.2.1'
        """

        # check flv_rna first. otherwise it may be considered as scopeV2.1.1 and scopeV2.2.1
        linker_flv_rna = Barcode.get_seq_str(seq, self.pattern_dict_flv_rna["L"])
        bool_valid, _, _ = Barcode.check_seq_mismatch(
            [linker_flv_rna], self.linker_flv_rna_set_list, self.linker_flv_rna_mismatch_list)
        if bool_valid:
            return "flv_rna"


        linker_v2 = Barcode.get_seq_str(seq, self.pattern_dict_v2["L"])
        bool_valid, _, _ = Barcode.check_seq_mismatch(
            [linker_v2], self.linker_1_v2_set_list, self.linker_1_v2_mismatch_list)
        if bool_valid:
            if seq[65:69] == "TTTT":
                return "scopeV2.0.1"
            else:
                return "scopeV2.1.1"

        linker_v3 = Barcode.get_seq_str(seq, self.pattern_dict_v3["L"])
        bool_valid, _, _ = Barcode.check_seq_mismatch(
            [linker_v3], self.linker_v3_set_list, self.linker_v3_mismatch_list)
        if bool_valid:
            return "scopeV3.0.1"

        linker_v2 = Barcode.get_seq_str(seq, self.pattern_dict_v2["L"])
        bool_valid, _, _ = Barcode.check_seq_mismatch(
            [linker_v2], self.linker_4_v2_set_list, self.linker_4_v2_mismatch_list)
        if bool_valid:
            return "scopeV2.2.1"

        linker_flv = Barcode.get_seq_str(seq, self.pattern_dict_flv["L"])
        bool_valid, _, _ = Barcode.check_seq_mismatch(
            [linker_flv], self.linker_flv_set_list, self.linker_flv_mismatch_list)
        if bool_valid:
            return "flv"

        return

    @utils.add_log
    def get_chemistry(self, fq1):
        results = defaultdict(int)

        with pysam.FastxFile(fq1) as fh:
            for _ in range(self.n_read):
                entry = fh.__next__()
                seq = entry.sequence
                chemistry = self.seq_chemistry(seq)
                if chemistry:
                    results[chemistry] += 1
        # if it is 0, then no other linker types
        if results["scopeV2.2.1"] != 0:
            results["scopeV2.2.1"] += results["scopeV2.1.1"]
        sorted_counts = sorted(results.items(), key=lambda x: x[1], reverse=True)
        self.get_chemistry.logger.info(sorted_counts)

        chemistry, read_counts = sorted_counts[0][0], sorted_counts[0][1]
        percent = float(read_counts) / self.n_read
        if percent < 0.5:
            self.get_chemistry.logger.warning("Valid chemistry read counts percent < 0.5")
        if percent < 0.1:
            self.get_chemistry.logger.error("Valid chemistry read counts percent < 0.1")
            raise Exception(
                'Auto chemistry detection failed! ' + HELP_DICT['chemistry']
            )
        Chemistry.get_chemistry.logger.info(f'chemistry: {chemistry}')

        return chemistry


class Barcode(Step):
    """
    ## Features
    - Demultiplex barcodes.
    - Filter invalid R1 reads, which includes:
        - Reads without linker: the mismatch between linkers and all linkers in the whitelist is greater than 2.  
        - Reads without correct barcode: the mismatch between barcodes and all barcodes in the whitelist is greater than 1.  
        - Reads without polyT: the number of T bases in the defined polyT region is less than 10.
        - Low quality reads: low sequencing quality in barcode and UMI regions.
    ## Output
    - `01.barcode/{sample}_2.fq(.gz)` Demultiplexed R2 reads. Barcode and UMI are contained in the read name. The format of 
    the read name is `{barcode}_{UMI}_{read ID}`.
    """

    def __init__(self, args, step_name):
        Step.__init__(self, args, step_name)

        self.fq1_list = args.fq1.split(",")
        self.fq2_list = args.fq2.split(",")
        self.fq_number = len(self.fq1_list)
        if self.fq_number != len(self.fq2_list):
            raise Exception('fastq1 and fastq2 do not have same file number!')
        if args.chemistry == 'auto':
            ch = Chemistry(args.fq1)
            self.chemistry_list = ch.check_chemistry()
        else:
            self.chemistry_list = [args.chemistry] * self.fq_number
        self.barcode_corrected_num = 0
        self.linker_corrected_num = 0
        self.total_num = 0
        self.clean_num = 0
        self.no_polyT_num = 0
        self.lowQual_num = 0
        self.no_linker_num = 0
        self.no_barcode_num = 0
        self.barcode_qual_Counter = Counter()
        self.umi_qual_Counter = Counter()
        self.pattern = args.pattern
        self.linker = args.linker
        self.whitelist = args.whitelist
        self.lowNum = args.lowNum
        self.lowQual = args.lowQual
        self.allowNoPolyT = args.allowNoPolyT
        self.allowNoLinker = args.allowNoLinker
        self.nopolyT = args.nopolyT  # true == output nopolyT reads
        self.noLinker = args.noLinker
        self.output_R1 = args.output_R1

        # out file
        if args.gzip:
            self.suffix = ".gz"
        else:
            self.suffix = ""
        self.out_fq2 = f'{self.out_prefix}_2.fq{self.suffix}'
        self.out_fq1 = f'{self.out_prefix}_1.fq{self.suffix}'
        if self.nopolyT:
            self.nopolyT_1 = f'{self.out_prefix}_noPolyT_1.fq'
            self.nopolyT_2 = f'{self.out_prefix}_noPolyT_2.fq'
        if self.noLinker:
            self.noLinker_1 = f'{self.out_prefix}_noLinker_1.fq'
            self.noLinker_2 = f'{self.out_prefix}_noLinker_2.fq'

        self.open_files()

    @staticmethod
    def get_seq_str(seq, sub_pattern_dict):
        """get subseq with intervals in arr and concatenate"""
        return ''.join([seq[item[0]: item[1]] for item in sub_pattern_dict])

    @staticmethod
    def get_seq_list(seq, pattern_dict, abbr):
        """
        >>> pattern_dict = Barcode.parse_pattern("C2L3C2")
        >>> seq = "AAGGGTT"
        >>> Barcode.get_seq_list(seq, pattern_dict, "C")
        ['AA', 'TT']
        """
        
        return [seq[item[0]: item[1]] for item in pattern_dict[abbr]]

    @staticmethod
    def parse_pattern(pattern):
        """
        >>> pattern_dict = Barcode.parse_pattern("C8L16C8L16C8L1U12T18")
        >>> pattern_dict['C']
        [[0, 8], [24, 32], [48, 56]]
        >>> pattern_dict['L']
        [[8, 24], [32, 48], [56, 57]]
        """
        pattern_dict = defaultdict(list)
        p = re.compile(r'([CLUNT])(\d+)')
        tmp = p.findall(pattern)
        if not tmp:
            Barcode.parse_pattern.logger.error(f'Invalid pattern: {pattern}')
            sys.exit()
        start = 0
        for item in tmp:
            end = start + int(item[1])
            pattern_dict[item[0]].append([start, end])
            start = end
        return pattern_dict

    @staticmethod
    def get_scope_bc(chemistry):
        """Return (linker file path, whitelist file path)"""
        try:
            linker_f = glob.glob(f'{ROOT_PATH}/data/chemistry/{chemistry}/linker*')[0]
            whitelist_f = f'{ROOT_PATH}/data/chemistry/{chemistry}/bclist'
        except IndexError:
            return None, None
        return linker_f, whitelist_f

    @staticmethod
    def ord2chr(q, offset=33):
        return chr(int(q) + offset)

    @staticmethod
    def qual_int(char, offset=33):
        return ord(char) - offset

    @staticmethod
    def low_qual(quals, minQ, num):
        # print(ord('/')-33)           14
        return True if len([q for q in quals if Barcode.qual_int(q) < minQ]) > num else False

    @staticmethod
    def findall_mismatch(seq, n_mismatch=1, bases='ACGTN'):
        """
        choose locations where there's going to be a mismatch using combinations
        and then construct all satisfying lists using product
        Return:
        all mismatch <= n_mismatch set. 
        >>> answer = set(["TCG", "AAG", "ACC", "ATG", "ACT", "ACN", "GCG", "ANG", "ACA", "ACG", "CCG", "AGG", "NCG"])
        >>> seq_set = Barcode.findall_mismatch("ACG")
        >>> seq_set == answer
        True
        """
        seq_set = set()
        seq_len = len(seq)
        if n_mismatch > seq_len:
            n_mismatch = seq_len
        for locs in combinations(range(seq_len), n_mismatch):
            seq_locs = [[base] for base in seq]
            for loc in locs:
                seq_locs[loc] = list(bases)
            for poss in product(*seq_locs):
                seq_set.add(''.join(poss))
        return seq_set

    @staticmethod
    @utils.add_log
    def get_mismatch_dict(seq_list, n_mismatch=1):
        """
        Return:
        mismatch dict. Key: mismatch seq, value: seq in seq_list
        >>> seq_list = ["AACGTGAT", "AAACATCG"]
        >>> mismatch_dict = Barcode.get_mismatch_dict(seq_list)
        >>> mismatch_dict["AACGTGAA"] == "AACGTGAT"
        True
        """
        mismatch_dict = {}

        for seq in seq_list:
            seq = seq.strip()
            if seq == '':
                continue
            for mismatch_seq in Barcode.findall_mismatch(seq, n_mismatch):
                mismatch_dict[mismatch_seq] = seq

        return mismatch_dict

    @staticmethod
    def check_seq_mismatch(seq_list, correct_set_list, mismatch_dict_list):
        '''
        Return bool_valid, bool_corrected, corrected_seq
        >>> seq_list = ['ATA', 'AAT', 'ATA']
        >>> correct_set_list = [{'AAA'},{'AAA'},{'AAA'}]
        >>> mismatch_dict_list = [Barcode.get_mismatch_dict(['AAA'])] * 3
        >>> Barcode.check_seq_mismatch(seq_list, correct_set_list, mismatch_dict_list)
        (True, True, 'AAAAAAAAA')
        >>> seq_list = ['AAA', 'AAA', 'AAA']
        >>> Barcode.check_seq_mismatch(seq_list, correct_set_list, mismatch_dict_list)
        (True, False, 'AAAAAAAAA')
        '''
        bool_valid = True
        bool_corrected = False
        corrected_seq = ''
        for index, seq in enumerate(seq_list):
            if seq not in correct_set_list[index]:
                if seq not in mismatch_dict_list[index]:
                    bool_valid = False
                    return bool_valid, bool_corrected, corrected_seq
                else:
                    bool_corrected = True
                    corrected_seq += mismatch_dict_list[index][seq]
            else:
                corrected_seq += seq
        return bool_valid, bool_corrected, corrected_seq

    @staticmethod
    def parse_whitelist_file(whitelist_file, n_mismatch, n_repeat):
        barcode_list, _ = utils.read_one_col(whitelist_file)
        barcode_set = set(barcode_list)
        barcode_mismatch_dict = Barcode.get_mismatch_dict(barcode_list, n_mismatch)
        barcode_mismatch_list = [barcode_mismatch_dict] * n_repeat
        barcode_set_list = [barcode_set] * n_repeat
        return barcode_set_list, barcode_mismatch_list

    @staticmethod
    def parse_linker_file(linker_file):
        return Barcode.parse_whitelist_file(linker_file, n_mismatch=2, n_repeat=1)

    @staticmethod
    def parse_chemistry(chemistry):
        """
        Returns: pattern_dict, barcode_set_list, barcode_mismatch_list, linker_set_list, linker_mismatch_list
        """
        pattern = PATTERN_DICT[chemistry]
        pattern_dict = Barcode.parse_pattern(pattern)
        linker_file, whitelist_file = Barcode.get_scope_bc(chemistry)

        barcode_set_list, barcode_mismatch_list = Barcode.parse_whitelist_file(whitelist_file, 1, len(pattern_dict['C']))
        linker_set_list, linker_mismatch_list = Barcode.parse_linker_file(linker_file)

        return pattern_dict, barcode_set_list, barcode_mismatch_list, linker_set_list, linker_mismatch_list

    @staticmethod
    def check_polyT(seq, pattern_dict, min_polyT_count=MIN_T):
        """
        Return:
            True if polyT is found
        """
        seq_polyT = Barcode.get_seq_str(seq, pattern_dict['T'])
        n_polyT_found = seq_polyT.count('T')
        if n_polyT_found >= min_polyT_count:
            return True
        return False

    def open_files(self):
        if self.output_R1:
            self.fh_fq1 = xopen(self.out_fq1, 'w')
        self.fh_fq2 = xopen(self.out_fq2, 'w')

        if self.nopolyT:
            self.fh_nopolyT_fq1 = xopen(self.nopolyT_1, 'w')
            self.fh_nopolyT_fq2 = xopen(self.nopolyT_2, 'w')

        if self.noLinker:
            self.fh_nolinker_fq1 = xopen(self.noLinker_1, 'w')
            self.fh_nolinker_fq2 = xopen(self.noLinker_2, 'w')

    def close_files(self):
        if self.output_R1:
            self.fh_fq1.close()
        self.fh_fq2.close()

        if self.nopolyT:
            self.fh_nopolyT_fq1.close()
            self.fh_nopolyT_fq2.close()
        
        if self.noLinker:
            self.fh_nolinker_fq1.close()
            self.fh_nolinker_fq2.close()

    @utils.add_log
    def add_step_metrics(self):

        self.add_metric(
            name='Raw Reads',
            value=self.total_num,
        )
        self.add_metric(
            name='Valid Reads',
            value=self.clean_num,
            total=self.total_num,
        )

        BarcodesQ30 = sum([self.barcode_qual_Counter[k] for k in self.barcode_qual_Counter if k >= self.ord2chr(
            30)]) / float(sum(self.barcode_qual_Counter.values())) * 100
        BarcodesQ30 = round(BarcodesQ30, 2)
        BarcodesQ30_display = f'{BarcodesQ30}%'
        self.add_metric(
            name='Q30 of Barcodes',
            value=BarcodesQ30_display,
        )

        UMIsQ30 = sum([self.umi_qual_Counter[k] for k in self.umi_qual_Counter if k >= self.ord2chr(
            30)]) / float(sum(self.umi_qual_Counter.values())) * 100
        UMIsQ30 = round(UMIsQ30, 2)
        UMIsQ30_display = f'{UMIsQ30}%'
        self.add_metric(
            name='Q30 of UMIs',
            value=UMIsQ30_display,
        )

        if self.clean_num == 0:
            raise Exception('no valid reads found! please check the --chemistry parameter.' + HELP_DICT['chemistry'])


    @utils.add_log
    def run(self):
        """
        Extract barcode and UMI from R1. Filter reads with 
            - invalid polyT
            - low quality in barcode and UMI
            - invalid inlinker
            - invalid barcode
            
        for every sample
            get chemistry
            get linker_mismatch_dict and barcode_mismatch_dict
            for every read in read1
                filter
                write valid R2 read to file
        """

        for i in range(self.fq_number):

            chemistry = self.chemistry_list[i]
            lowNum = int(self.lowNum)
            lowQual = int(self.lowQual)
            if chemistry == 'scopeV1':
                lowNum = min(0, lowNum)
                lowQual = max(10, lowQual)
                self.run.logger.info(f'scopeV1: lowNum={lowNum}, lowQual={lowQual} ')
            # get linker and whitelist
            bc_pattern = PATTERN_DICT[chemistry]
            if (bc_pattern):
                linker_file, whitelist_file = self.get_scope_bc(chemistry)
            else:
                bc_pattern = self.pattern
                linker_file = self.linker
                whitelist_file = self.whitelist
            if not bc_pattern:
                raise Exception("invalid bc_pattern!")

            pattern_dict = self.parse_pattern(bc_pattern)

            bool_T = True if 'T' in pattern_dict else False
            bool_L = True if 'L' in pattern_dict else False
            bool_whitelist = (whitelist_file is not None) and whitelist_file != "None"
            C_len = sum([item[1] - item[0] for item in pattern_dict['C']])

            if bool_whitelist:
                barcode_set_list, barcode_mismatch_list = Barcode.parse_whitelist_file(whitelist_file,
                                                                               n_mismatch=1, n_repeat=len(pattern_dict['C']))
            if bool_L:
                linker_set_list, linker_mismatch_list = Barcode.parse_linker_file(linker_file)

            with pysam.FastxFile(self.fq1_list[i], persist=False) as fq1, \
                    pysam.FastxFile(self.fq2_list[i], persist=False) as fq2:
                for entry1, entry2 in zip(fq1, fq2):
                    header1, seq1, qual1 = entry1.name, entry1.sequence, entry1.quality
                    header2, seq2, qual2 = entry2.name, entry2.sequence, entry2.quality
                    self.total_num += 1

                    # polyT filter
                    if bool_T and (not self.allowNoPolyT):
                        if not Barcode.check_polyT(seq1, pattern_dict):
                            self.no_polyT_num += 1
                            if self.nopolyT:
                                self.fh_nopolyT_fq1.write(
                                    '@%s\n%s\n+\n%s\n' % (header1, seq1, qual1))
                                self.fh_nopolyT_fq2.write(
                                    '@%s\n%s\n+\n%s\n' % (header2, seq2, qual2))
                            continue

                    # lowQual filter
                    C_U_quals_ascii = Barcode.get_seq_str(
                        qual1, pattern_dict['C'] + pattern_dict['U'])
                    if lowQual > 0 and Barcode.low_qual(C_U_quals_ascii, lowQual, lowNum):
                        self.lowQual_num += 1
                        continue

                    # linker filter
                    if bool_L and (not self.allowNoLinker):
                        seq_str = Barcode.get_seq_str(seq1, pattern_dict['L'])
                        bool_valid, bool_corrected, _ = Barcode.check_seq_mismatch(
                            [seq_str], linker_set_list, linker_mismatch_list)
                        if not bool_valid:
                            self.no_linker_num += 1
                            if self.noLinker:
                                self.fh_noLinker_fq1(
                                    '@%s\n%s\n+\n%s\n' % (header1, seq1, qual1))
                                self.fh_noLinker_fq2(
                                    '@%s\n%s\n+\n%s\n' % (header2, seq2, qual2))
                            continue
                        elif bool_corrected:
                            self.linker_corrected_num += 1

                    # barcode filter
                    seq_list = self.get_seq_list(seq1, pattern_dict, 'C')
                    if bool_whitelist:
                        bool_valid, bool_corrected, corrected_seq = Barcode.check_seq_mismatch(
                            seq_list, barcode_set_list, barcode_mismatch_list)

                        if not bool_valid:
                            self.no_barcode_num += 1
                            continue
                        elif bool_corrected:
                            self.barcode_corrected_num += 1
                        cb = corrected_seq
                    else:
                        cb = "".join(seq_list)

                    self.clean_num += 1
                    self.barcode_qual_Counter.update(C_U_quals_ascii[:C_len])
                    self.umi_qual_Counter.update(C_U_quals_ascii[C_len:])

                    umi = Barcode.get_seq_str(seq1, pattern_dict['U'])
                    self.fh_fq2.write(f'@{cb}_{umi}_{self.total_num}\n{seq2}\n+\n{qual2}\n')
                    if self.output_R1:
                        self.fh_fq1.write(f'@{cb}_{umi}_{self.total_num}\n{seq1}\n+\n{qual1}\n')                   
            
            self.run.logger.info(self.fq1_list[i] + ' finished.')

        self.close_files()
        self.add_step_metrics()
        self.clean_up()


@utils.add_log
def barcode(args):
    step_name = "barcode"
    obj = Barcode(args, step_name)
    obj.run()


def get_opts_barcode(parser, sub_program=True):
    parser.add_argument(
        '--chemistry',
        help='Predefined (pattern, barcode whitelist, linker whitelist) combinations. ' + HELP_DICT['chemistry'],
        choices=list(PATTERN_DICT.keys()),
        default='auto'
    )
    parser.add_argument(
        '--pattern',
        help="""The pattern of R1 reads, e.g. `C8L16C8L16C8L1U12T18`. The number after the letter represents the number 
        of bases.  
- `C`: cell barcode  
- `L`: linker(common sequences)  
- `U`: UMI    
- `T`: poly T""",
    )
    parser.add_argument(
        '--whitelist',
        help='Cell barcode whitelist file path, one cell barcode per line.'
    )
    parser.add_argument(
        '--linker',
        help='Linker whitelist file path, one linker per line.'
    )
    parser.add_argument(
        '--lowQual',
        help='Default 0. Bases in cell barcode and UMI whose phred value are lower than \
lowQual will be regarded as low-quality bases.',
        type=int,
        default=0
    )
    parser.add_argument(
        '--lowNum',
        help='The maximum allowed lowQual bases in cell barcode and UMI.',
        type=int,
        default=2
    )
    parser.add_argument(
        '--nopolyT',
        help='Outputs R1 reads without polyT.',
        action='store_true',
    )
    parser.add_argument(
        '--noLinker',
        help='Outputs R1 reads without correct linker.',
        action='store_true',
    )
    parser.add_argument(
        '--allowNoPolyT',
        help="Allow valid reads without polyT.",
        action='store_true'
    )
    parser.add_argument(
        '--allowNoLinker',
        help="Allow valid reads without correct linker.",
        action='store_true'
    )
    parser.add_argument(
        '--gzip',
        help="Output gzipped fastq files.",
        action='store_true'
    )
    parser.add_argument(
        '--output_R1',
        help="Output valid R1 reads.",
        action='store_true'
    )
    if sub_program:
        parser.add_argument('--fq1', help='R1 fastq file. Multiple files are separated by comma.', required=True)
        parser.add_argument('--fq2', help='R2 fastq file. Multiple files are separated by comma.', required=True)
        parser = s_common(parser)

    return parser


if __name__ == '__main__':
    unittest.main()


import abc
import collections
import io
import json
import numbers
import os
from collections import namedtuple

import pandas as pd
from jinja2 import Environment, FileSystemLoader, select_autoescape

from celescope.tools.utils import add_log
from celescope.__init__ import HELP_DICT


Metric = namedtuple("Metric", "name value total fraction")


def s_common(parser):
    """subparser common arguments
    """
    parser.add_argument('--outdir', help='Output diretory.', required=True)
    parser.add_argument('--assay', help='Assay name.', required=True)
    parser.add_argument('--sample', help='Sample name.', required=True)
    parser.add_argument('--thread', help=HELP_DICT['thread'], default=4)
    parser.add_argument('--debug', help=HELP_DICT['debug'], action='store_true')
    return parser


class Step:
    """
    Step class
    """

    def __init__(self, args, step_name):
        self.step_name = step_name
        self.args = args
        self.outdir = args.outdir
        self.sample = args.sample
        self.assay = args.assay
        self.thread = int(args.thread)
        self.debug = args.debug
        # set
        self.out_prefix = f'{self.outdir}/{self.sample}'

        # important! make outdir before path_dict because path_dict use relative path.
        if not os.path.exists(self.outdir):
            os.system('mkdir -p %s' % self.outdir)

        self.metric_list = []
        self.path_dict = {
            "metric": f'{self.outdir}/../.metrics.json',
            "data": f'{self.outdir}/../.data.json'
        }
        self.content_dict = {}
        for slot, path in self.path_dict.items():
            if not os.path.exists(path):
                self.content_dict[slot] = {}
            else:
                with open(path) as f:
                    self.content_dict[slot] = json.load(f)

        # jinja env
        self.env = Environment(
            loader=FileSystemLoader(os.path.dirname(__file__) + '/../templates/'),
            autoescape=select_autoescape(['html', 'xml'])
        )

        # out file
        self.stat_file = f'{self.outdir}/stat.txt'

    def add_metric(self, name, value=None, total=None, fraction=None):
        '''add metric to metric_list
        '''
        self.metric_list.append(Metric(
            name=name, value=value, total=total, fraction=fraction
        ))

    def get_fraction(self):
        '''
        metric_list: list of namedtuple(name, value, total, fraction)
        '''
        metric_list = []
        for metric in self.metric_list:
            fraction = metric.fraction
            if metric.total:
                fraction = metric.value / metric.total
            if fraction:
                fraction = round(fraction, 4)
            metric_list.append(Metric(
                name=metric.name,
                value=metric.value,
                total=metric.total,
                fraction=fraction,
            ))
        self.metric_list = metric_list

    def metric_list_to_stat(self):
        with open(self.stat_file, 'w') as stat_handle:
            for metric in self.metric_list:
                line = f'{metric.name}: '
                value = metric.value
                fraction = metric.fraction
                value_bool = value == 0 or value
                fraction_bool = fraction == 0 or fraction
                if fraction_bool:
                    fraction = round(fraction * 100, 2)
                if value_bool:
                    if isinstance(value, numbers.Number):
                        line += format(value, ',')
                        if fraction_bool:
                            line += f'({fraction}%)'
                    else:
                        line += value
                elif fraction_bool:
                    line += f'{fraction}%'
                stat_handle.write(line + '\n')

    def dump_content(self, slot):
        '''dump content to json file
        '''
        if self.content_dict[slot]:
            with open(self.path_dict[slot], 'w') as f:
                json.dump(self.content_dict[slot], f, indent=4)

    @add_log
    def render_html(self):
        template = self.env.get_template(f'html/{self.assay}/base.html')
        report_html = f"{self.outdir}/../{self.sample}_report.html"
        with io.open(report_html, 'w', encoding='utf8') as f:
            html = template.render(self.content_dict['data'])
            f.write(html)

    def stat_to_data(self):
        df = pd.read_table(self.stat_file, header=None, sep=':', dtype=str)
        self.content_dict['data'][self.step_name + '_summary'] = df.values.tolist()

    def stat_to_metric(self):
        '''
        can be:
        1. value
        2. value(fraction%)
        3. fraction%
        '''

        df = pd.read_table(self.stat_file, header=None, sep=':', dtype=str)
        dic = dict(zip(df.iloc[:, 0], df.iloc[:, 1].str.strip()))
        metrics = dict()
        for metric_name, string in dic.items():
            bool_fraction = False
            bool_value = False
            if '%' in string:
                bool_fraction = True
                if "(" in string:
                    bool_value = True
            chars = [',', '%', ')']
            for character in chars:
                string = string.replace(character, '')

            if bool_fraction:
                if bool_value:  # case 2
                    value, fraction = string.split('(')
                    fraction = round(float(fraction) / 100, 4)
                    metrics[metric_name] = int(value)
                    metrics[metric_name + ' Fraction'] = fraction
                else:  # case 3
                    fraction = round(float(string) / 100, 4)
                    metrics[metric_name] = fraction
            else:  # case 1
                value = string
                if '.' in string:
                    try:
                        value = round(float(string), 4)
                    except ValueError:
                        pass
                else:
                    try:
                        value = int(string)
                    except ValueError:
                        pass
                metrics[metric_name] = value

        self.content_dict['metric'][self.step_name + '_summary'] = metrics

    def add_content_item(self, slot, **kwargs):
        for key, value in kwargs.items():
            # if value is a dict, and some value in this dict is float, format these value
            if isinstance(value, collections.abc.Mapping):
                for value_key, value_value in value.items():
                    if isinstance(value_value, float):
                        value[value_key] = round(value_value, 4)

            self.content_dict[slot][key] = value

    def add_data_item(self, **kwargs):
        self.add_content_item("data", **kwargs)

    @staticmethod
    def get_table(title, table_id, df_table):
        """
        return html code
        """
        table_dict = {}
        table_dict['title'] = title
        table_dict['table'] = df_table.to_html(
            escape=False,
            index=False,
            table_id=table_id,
            justify="center")
        table_dict['id'] = table_id
        return table_dict

    @add_log
    def clean_up(self):
        if self.metric_list:
            self.get_fraction()
            self.metric_list_to_stat()
        if os.path.exists(self.stat_file):
            self.stat_to_metric()
            self.stat_to_data()
        self.dump_content(slot="data")
        self.dump_content(slot="metric")
        self.render_html()

    @abc.abstractmethod
    def run(self):
        return



from celescope.dynaseq.__init__ import __ASSAY__
from celescope.tools.multi import Multi


class Multi_dynaseq(Multi):

    def conversion(self, sample):
        step = 'conversion'
        bam = f'{self.outdir_dic[sample]["featureCounts"]}/{sample}_Aligned.sortedByCoord.out.bam.featureCounts.bam'
        cell = f'{self.outdir_dic[sample]["count"]}/{sample}_matrix_10X/barcodes.tsv'
        cmd_line = self.get_cmd_line(step, sample)
        cmd = (
            f'{cmd_line} '
            f'--bam {bam} '
            f'--cell {cell} '
        )
        self.process_cmd(cmd, step, sample, m=5, x=1)

    def substitution(self, sample):
        step = 'substitution'
        bam = f'{self.outdir_dic[sample]["conversion"]}/{sample}.PosTag.bam'
        cmd_line = self.get_cmd_line(step, sample)
        cmd = (
            f'{cmd_line} '
            f'--bam {bam} '
        )
        self.process_cmd(cmd, step, sample, m=1, x=1)


    def replacement(self, sample):
        step = 'replacement'
        bam = f'{self.outdir_dic[sample]["conversion"]}/{sample}.PosTag.bam'
        cmd_line = self.get_cmd_line(step, sample)
        cmd = (
            f'{cmd_line} '
            f'--bam {bam} '
            f'--bg {self.col5_dict[sample]} '
        )
        self.process_cmd(cmd, step, sample, m=10, x=1)
    

    def replace_tsne(self, sample):
        step = 'replace_tsne'
        tsne_file = f'{self.outdir_dic[sample]["analysis"]}/{sample}_tsne_coord.tsv'
        mat_file = f'{self.outdir_dic[sample]["replacement"]}/{sample}.fraction_of_newRNA_matrix.txt'
        rep_file = f'{self.outdir_dic[sample]["replacement"]}/{sample}.fraction_of_newRNA_per_cell.txt'
        cmd_line = self.get_cmd_line(step, sample)
        cmd = (
            f'{cmd_line} '
            f'--tsne {tsne_file} '
            f'--mat {mat_file} '
            f'--rep {rep_file} '
        )
        self.process_cmd(cmd, step, sample, m=1, x=1)


def main():
    multi = Multi_dynaseq(__ASSAY__)
    multi.run()

if __name__ == '__main__':
    main()




import argparse
import json
import os
from collections import defaultdict

import celescope.tools.utils as utils


@utils.add_log
def merge_report():
    parser = argparse.ArgumentParser('merge report')
    parser.add_argument('--outdir', help='outdir', required=True)
    parser.add_argument(
        '--samples', help='samples, seperated by comma', required=True)
    parser.add_argument('--steps', help='steps', required=True)
    parser.add_argument('--rm_files', action='store_true',
                        help='remove all fq and bam after running')
    args = vars(parser.parse_args())

    outdir = args['outdir']
    os.chdir(outdir)
    steps = args['steps'].split(",")
    samples = args['samples'].split(',')
    result_dict = defaultdict(list)

    summarys = [step + '_summary' for step in steps]
    for sample in samples:
        data_json = f'./{sample}/.data.json'
        data_dic = json.load(open(data_json))
        for summary in summarys:
            if summary not in data_dic.keys():
                # TCR and BCR data_dict has prefix
                bool_in = False
                for key in data_dic.keys():
                    if key.find(summary) != -1:
                        summary = key
                        bool_in = True
                if not bool_in:
                    continue

            # add title
            if sample == samples[0]:
                result_dict[summary].append(
                    '\t'.join([str(x[0]).replace(' ', '_') for x in data_dic[summary]])
                )
            result_dict[summary].append(
                '\t'.join([str(x[1]).replace(' ', '') for x in data_dic[summary]])
            )

    with open('./merge.xls', 'w') as fh:
        for summary in result_dict.keys():
            fh.write('##' + summary + '\n')
            for k in result_dict[summary]:
                fh.write(k + '\n')
            fh.write('\n')

    if args['rm_files']:
        rm_files()


@utils.add_log
def rm_files():
    cmd = '''
        find . -iname '*.fq*' -delete;
        find . -iname '*.bam' -not -path './*/*.featureCounts/*name_sorted.bam' -delete;
    '''
    os.system(cmd)


if __name__ == '__main__':
    merge_report()



import abc
import configparser

from celescope.tools.__init__ import GENOME_CONFIG


def parse_genomeDir(genomeDir, entrys=None):
    config_file = f'{genomeDir}/{GENOME_CONFIG}'
    config = configparser.ConfigParser()
    config.read_file(open(config_file))
    genome = config['genome']
    # add path
    if entrys:
        for entry in entrys:
            if entry in genome and genome[entry] != 'None':
                genome[entry] = f'{genomeDir}/{genome[entry]}'
            else:
                genome[entry] = "None"
    return genome


class Mkref():

    def __init__(self, genome_type, args):
        self.genomeDir = args.genomeDir
        self.thread = args.thread
        self.genome_name = args.genome_name
        self.dry_run = args.dry_run
        self.genome_type = genome_type

        # out file
        self.config_file = f'{self.genomeDir}/{GENOME_CONFIG}'

    @abc.abstractmethod
    def run(self):
        return

    @abc.abstractmethod
    def write_config(self):
        return


def get_opts_mkref(parser, sub_program):
    if sub_program:
        parser.add_argument("--genomeDir", help="Default='./'. Output directory.", default='./')
        parser.add_argument("--thread", help="Default=6. Threads to use.", default=6)
        parser.add_argument("--genome_name", help="Required, genome name. ", required=True)
        parser.add_argument("--dry_run", help="Only write config file and exit.", action='store_true')



"""
split scRNA-Seq fastq file(01.barcode/{sample}_2.fq)
"""
import glob
import os
from collections import defaultdict

import pysam
import pandas as pd

import celescope.tools.utils as utils
from celescope.tools.step import Step, s_common
from celescope.__init__ import HELP_DICT
from celescope.tools.cellranger3.wrapper import Cell_calling, read_raw_matrix


def get_clonotypes_table(df):
    chains = sorted(set(df['chain'].tolist()))
    res = pd.DataFrame(columns=['barcode'])
    for c in chains:
        df_c = df[df['chain']==c][['barcode', 'aaSeqCDR3', 'nSeqCDR3']]
        df_c = df_c.rename(columns={'aaSeqCDR3': f'{c}_aaSeqCDR3', 'nSeqCDR3': f'{c}_nSeqCDR3'})
        res = pd.merge(res, df_c, on='barcode', how='outer').fillna('NaN')
    group_l = res.columns.tolist()
    group_l.remove('barcode')
    clonotypes = res.groupby(group_l, as_index=False).agg({'barcode': 'count'})
    clonotypes = clonotypes.rename(columns={'barcode': 'barcode_count'})
    sum_cb = clonotypes['barcode_count'].sum()
    clonotypes['percent'] = clonotypes['barcode_count'].apply(lambda x: round(x/sum_cb, 2))
    clonotypes['clonetype_ID'] = [i for i in range(1, clonotypes.shape[0]+1)]
    group_l.insert(0, 'clonetype_ID')
    group_l.append('barcode_count')
    group_l.append('percent')
    clonotypes = clonotypes.reindex(columns=group_l)
    if chains[0].startswith("TR"):
        return clonotypes, 'TCR'
    elif chains[0].startswith("IG"):
        return clonotypes, 'BCR'


class Split_tag(Step):
    """
    Features
    - Split scRNA-Seq fastq according to tag assignment.

    Output
    - `matrix/` Matrix files of each tag.(Optional)
    - `fastq/` Fastq files of each tag.(Optional)
    """

    def __init__(self, args, step_name):
        Step.__init__(self, args, step_name)
        if not (args.split_matrix or args.split_fastq or args.split_vdj):
            return

        # set
        df_umi_tag = pd.read_csv(args.umi_tag_file, sep='\t', index_col=0)
        df_umi_tag = df_umi_tag.rename_axis('barcode').reset_index()
        self.tag_barcode_dict = {tag: set(row["barcode"].tolist()) for tag, row in df_umi_tag.groupby("tag")}

        if args.split_matrix:
            self.matrix_outdir = f'{args.outdir}/matrix/'
            if args.match_dir:
                matrix_10X_dir = glob.glob(f'{args.match_dir}/05.count/*_matrix_10X*')[0]
            elif args.matrix_dir:
                matrix_10X_dir = args.matrix_dir
            else:
                raise ValueError("--match_dir or --matrix_dir is required.")
            self.raw_mat, self.raw_features_path, self.raw_barcodes = read_raw_matrix(matrix_10X_dir)

        if args.split_fastq:
            self.rna_fq_file = glob.glob(f'{args.match_dir}/*barcode/*_2.fq*')[0]

            fastq_outdir = f'{args.outdir}/fastqs/'
            os.system(f'mkdir -p {fastq_outdir}')

            self.r2_fastq_files_handle = {}
            self.r1_fastq_files_handle = {}
            for tag in self.tag_barcode_dict:
                r2_fastq_file_name = f'{fastq_outdir}/{tag}_2.fq'
                self.r2_fastq_files_handle[tag] = open(r2_fastq_file_name, 'w')
                r1_fastq_file_name = f'{fastq_outdir}/{tag}_1.fq'
                self.r1_fastq_files_handle[tag] = open(r1_fastq_file_name, 'w')

            self.tag_read_index_dict = defaultdict(set)
        
        if args.split_vdj:
            self.cell_confident_vdj = glob.glob(f'{args.vdj_dir}/*count_vdj/*cell_confident.tsv*')[0]

            self.vdj_outdir = f'{args.outdir}/vdj/'
            if not os.path.exists(self.vdj_outdir):
                os.system(f'mkdir -p {self.vdj_outdir}')
            

    @utils.add_log
    def write_r2_fastq_files(self):
        read_num = 0
        with pysam.FastxFile(self.rna_fq_file, 'r') as rna_fq:
            for read in rna_fq:
                read_num += 1
                attr = read.name.strip("@").split("_")
                barcode = attr[0]
                read_index = int(attr[2])
                for tag in self.tag_barcode_dict:
                    if barcode in self.tag_barcode_dict[tag]:
                        self.tag_read_index_dict[tag].add(read_index)
                        self.r2_fastq_files_handle[tag].write(str(read) + '\n')

                if read_num % 1000000 == 0:
                    self.write_r2_fastq_files.logger.info(f'{read_num} done')

        for tag in self.r2_fastq_files_handle:
            self.r2_fastq_files_handle[tag].close()

    @utils.add_log
    def write_r1_fastq_files(self):
        with pysam.FastxFile(self.args.R1_read, 'r') as r1_read:
            for read_index, read in enumerate(r1_read, start=1):
                for tag in self.tag_read_index_dict:
                    if read_index in self.tag_read_index_dict[tag]:
                        self.r1_fastq_files_handle[tag].write(str(read) + '\n')

        for tag in self.r1_fastq_files_handle:
            self.r1_fastq_files_handle[tag].close()

    @utils.add_log
    def split_matrix(self):
        for tag in self.tag_barcode_dict:
            outdir = f'{self.matrix_outdir}/{tag}_matrix_10X/'
            runner = Cell_calling(outdir, self.raw_mat, self.raw_features_path, self.raw_barcodes)
            tag_barcodes = list(self.tag_barcode_dict[tag])
            raw_barcodes = list(runner.raw_barcodes)
            tag_barcodes_indices = [raw_barcodes.index(barcode) for barcode in tag_barcodes]
            tag_barcodes_indices.sort()
            runner.write_slice_matrix(tag_barcodes_indices)

    @utils.add_log
    def split_vdj(self):

        df_vdj = pd.read_csv(self.cell_confident_vdj, sep='\t')
        for tag in self.tag_barcode_dict:
            tag_barcodes = list(self.tag_barcode_dict[tag])
            temp = df_vdj[df_vdj.barcode.isin(tag_barcodes)]
            if not temp.empty:
                clonotypes, seqtype = get_clonotypes_table(temp)
                temp.to_csv(f'{self.vdj_outdir}/{tag}_{seqtype}_cell_confident.tsv', sep='\t', index=False)
                clonotypes.to_csv(f'{self.vdj_outdir}/{tag}_{seqtype}_clonotypes.tsv', sep='\t', index=False)
            else:
                continue
            

    @utils.add_log
    def run(self):
        if self.args.split_matrix:
            self.split_matrix()
        if self.args.split_fastq:
            self.write_r2_fastq_files()
            self.write_r1_fastq_files()
        if self.args.split_vdj:
            self.split_vdj()


def split_tag(args):
    step_name = "split_tag"
    runner = Split_tag(args, step_name)
    runner.run()


def get_opts_split_tag(parser, sub_program):
    parser.add_argument(
        "--split_fastq",
        help="If used, will split scRNA-Seq fastq file according to tag assignment.",
        action='store_true',
    )
    parser.add_argument(
        "--split_matrix",
        help="If used, will split scRNA-Seq matrix file according to tag assignment.",
        action='store_true',
    )
    parser.add_argument(
        "--split_vdj", 
        help="If used, will split scRNA-Seq vdj count file according to tag assignment.", 
        action='store_true',
    )
    parser.add_argument("--vdj_dir", help="Match celescope vdj directory. Required when --split_vdj is specified.")
    if sub_program:
        parser.add_argument("--umi_tag_file", help="UMI tag file.", required=True)
        parser.add_argument("--match_dir", help=HELP_DICT['match_dir'])
        parser.add_argument("--matrix_dir", help="Match celescope scRNA-Seq matrix directory.")
        parser.add_argument("--R1_read", help='R1 read path.')
        s_common(parser)



import re
import subprocess
from itertools import islice

import pandas as pd
import pysam

from celescope.tools.step import Step, s_common
import celescope.tools.utils as utils

ADAPTER = ['polyT=A{18}', 'p5=AGATCGGAAGAGCACACGTCTGAACTCCAGTCAC']


class Cutadapt(Step):
    """
    Features
    - Trim adapters in R2 reads with cutadapt. Default adapters includes:
        - polyT=A{18}, 18 A bases. 
        - p5=AGATCGGAAGAGCACACGTCTGAACTCCAGTCA, Illumina p5 adapter.

    Output
    - `cutadapt.log` Cutadapt output log file.
    - `{sample}_clean_2.fq.gz` R2 reads file without adapters.
    """

    def __init__(self, args, step_name):
        Step.__init__(self, args, step_name)

        # set
        self.adapter_args = self.read_adapter_fasta(args.adapter_fasta)
        self.adapter_args += ADAPTER

        # out files
        if args.gzip:
            suffix = ".gz"
        else:
            suffix = ""
        self.out_fq2 = f'{self.outdir}/{self.sample}_clean_2.fq{suffix}'
        self.cutadapt_log_file = f'{self.outdir}/cutadapt.log'

    @staticmethod
    def read_adapter_fasta(adapter_fasta):
        '''
        return ['adapter1=AAA','adapter2=BBB']
        '''
        adapter_args = []
        if adapter_fasta and adapter_fasta != 'None':
            with pysam.FastxFile(adapter_fasta) as fh:
                for read in fh:
                    adapter_args.append(f'{read.name}={read.sequence}')
        return adapter_args

    def format_and_write_stat(self, cutadapt_log):
        # Total reads processed:...Total written (filtered):
        start_line = 9
        end_line = 16
        useful_content = islice(cutadapt_log.split('\n'), start_line, end_line)
        p_list = []
        for line_index, line in enumerate(useful_content):
            if line.strip() == '':
                continue
            line = re.sub(r'\s{2,}', r'', line)
            line = re.sub(r' bp', r'', line)
            line = re.sub(r'(?<=\d)\s+\(', r'(', line)
            line = line.strip()
            attr = line.split(":")
            if len(attr) < 2:
                raise Exception(
                    'May caused by cutadapt version != 1.1.7. '
                    f'Check version in {self.outdir}/cutadapt.log\n'
                    f'cutadapt log error at line {line_index + start_line}:\n'
                    f'{line}'
                )
            p_list.append({"item": attr[0], "value": attr[1]})
        p_df = pd.DataFrame(p_list)
        p_df.iloc[0, 0] = 'Reads with Adapters'
        p_df.iloc[1, 0] = 'Reads too Short'
        p_df.iloc[2, 0] = 'Reads Written'
        p_df.iloc[3, 0] = 'Base Pairs Processed'
        p_df.iloc[4, 0] = 'Base Pairs Quality-Trimmed'
        p_df.iloc[5, 0] = 'Base Pairs Written'
        p_df.to_csv(self.stat_file, sep=':', index=False, header=None)

    @utils.add_log
    def run(self):
        adapter_args_str = " ".join(['-a ' + adapter for adapter in self.adapter_args])
        cmd = (
            'cutadapt '
            f'{adapter_args_str} '
            f'-n {len(self.adapter_args)} '
            f'-j {self.thread} '
            f'-m {self.args.minimum_length} '
            f'--nextseq-trim={self.args.nextseq_trim} '
            f'--overlap {self.args.overlap} '
            f'-l {self.args.insert} '
            f'-o {self.out_fq2} '
            f'{self.args.fq} '
        )
        Cutadapt.run.logger.info(cmd)
        # need encoding argument to return str
        results = subprocess.run(
            cmd, stderr=subprocess.STDOUT, stdout=subprocess.PIPE,
            encoding='utf-8', check=True, shell=True
        )
        cutadapt_log = results.stdout
        with open(self.cutadapt_log_file, 'w') as f:
            f.write(cutadapt_log)
        self.format_and_write_stat(cutadapt_log)
        self.clean_up()


@utils.add_log
def cutadapt(args):

    step_name = "cutadapt"
    cutadapt_obj = Cutadapt(args, step_name)
    cutadapt_obj.run()


def get_opts_cutadapt(parser, sub_program):
    parser.add_argument('--gzip', help="Output gzipped fastq files.", action='store_true')
    parser.add_argument('--adapter_fasta', help='Addtional adapter fasta file.')
    parser.add_argument(
        '--minimum_length',
        help='Default `20`. Discard processed reads that are shorter than LENGTH.',
        default=20
    )
    parser.add_argument(
        '--nextseq_trim',
        help="""Default `20`. Quality trimming of reads using two-color chemistry (NextSeq). 
Some Illumina instruments use a two-color chemistry to encode the four bases. 
This includes the NextSeq and the NovaSeq. 
In those instruments, a ‘dark cycle’ (with no detected color) encodes a G. 
However, dark cycles also occur when sequencing “falls off” the end of the fragment.
The read then contains a run of high-quality, but incorrect “G” calls at its 3’ end.""",
        default=20,
    )
    parser.add_argument(
        '--overlap',
        help="""Default `10`. Since Cutadapt allows partial matches between the read and the adapter sequence,
short matches can occur by chance, leading to erroneously trimmed bases. 
For example, roughly 0.25 of all reads end with a base that is identical to the first base of the adapter. 
To reduce the number of falsely trimmed bases, the alignment algorithm requires that 
at least {overlap} bases match between adapter and read. """,
        default=10
    )
    parser.add_argument(
        '--insert',
        help="Default `150`. Read2 insert length.",
        default=150
    )
    if sub_program:
        parser.add_argument('--fq', help='Required. R2 reads from step Barcode.', required=True)
        parser = s_common(parser)
    return parser



__STEPS__ = [
    'mkref',
    'sample',
    'barcode',
    'cutadapt',
    'star',
    "featureCounts",
    "count",
    'analysis']
__ASSAY__ = 'rna'

# m: memory
# x: thread
RESOURCE = {
    'sample': {'m': 1, 'x': 1},
    'barcode': {'m': 5, 'x': 1},
    'cutadapt': {'m': 5, 'x': 1},
    'star': {'m': 30, 'x': 1},
}



'''
vdj mapping
'''

import subprocess

import pandas as pd
import pysam

import celescope.tools.utils as utils
from celescope.tools.step import Step, s_common
from celescope.vdj.__init__ import CHAINS


class Mapping_vdj(Step):
    """
    Features
    - Align R2 reads to IGMT(http://www.imgt.org/) database sequences with mixcr.

    Output
    - `{sample}_consensus.fasta` Fasta file after UMI consensus.

    - `{sample}_UMI_count_unfiltered.tsv` UMI reading for each (barcode, chain, VJ_pair) combination.

    - `{sample}_UMI_count_filtered.tsv` For each (barcode, chain) combination, only the record with the 
    most VJ_pair UMI reads is kept.

    - `{sample}_align.txt` Result report.

    - `{sample}_alignments.txt` The alignment result of each UMI/read.
    """

    def __init__(self, args, step_name):
        Step.__init__(self, args, step_name)

        # set
        self.read_type = "UMIs"
        if args.not_consensus:
            self.read_type = 'Reads'
        self.chains = CHAINS[args.type]

        # out files
        self.UMI_count_unfiltered_file = f'{self.out_prefix}_UMI_count_unfiltered.tsv'
        self.UMI_count_filtered_file = f'{self.out_prefix}_UMI_count_filtered.tsv'
        self.mixcr_report = f"{self.out_prefix}_align.txt"
        self.not_align_fq = f"{self.out_prefix}_not_align.fq"
        self.read2_vdjca = f"{self.out_prefix}_read2.vdjca"
        self.alignments = f"{self.out_prefix}_alignments.txt"

    @utils.add_log
    def run_mixcr(self):
        cmd = (
            'mixcr align '
            '--force-overwrite '
            f'--species {self.args.species} '
            f'-t {self.args.thread} '
            f'--not-aligned-R1 {self.not_align_fq} '
            f'--report {self.mixcr_report} '
            '-OallowPartialAlignments=true '
            '-OvParameters.geneFeatureToAlign=VTranscriptWithP '
            f'{self.args.fq} {self.read2_vdjca} '
            '\n'
            'mixcr exportAlignments '
            f'{self.read2_vdjca} {self.alignments} '
            '-readIds --force-overwrite -vGene -dGene -jGene -cGene '
            '-nFeature CDR3 -aaFeature CDR3 '
        )

        Mapping_vdj.run_mixcr.logger.info(cmd)
        subprocess.check_call(cmd, shell=True)

    @utils.add_log
    def mixcr_summary(self, total_read, df_align):

        align_read = df_align.shape[0]
        self.add_metric(
            name=f"{self.read_type} Mapped to Any VDJ Gene",
            value=align_read,
            total=total_read,
        )

        # CDR3
        df_CDR3 = df_align[~pd.isnull(df_align["aaSeqCDR3"])]
        align_read_with_CDR3 = df_CDR3.shape[0]
        self.add_metric(
            name=f"{self.read_type} with CDR3",
            value=align_read_with_CDR3,
            total=total_read,
        )

        # correct CDR3
        df_correct_CDR3 = df_CDR3[~(df_CDR3["aaSeqCDR3"].str.contains(r"\*"))]
        align_read_with_correct_CDR3 = df_correct_CDR3.shape[0]
        self.add_metric(
            name=f"{self.read_type} with Correct CDR3",
            value=align_read_with_correct_CDR3,
            total=total_read,
        )

        # VDJ
        df_VJ = df_correct_CDR3[
            (~pd.isnull(df_correct_CDR3['bestVGene'])) &
            (~pd.isnull(df_correct_CDR3['bestJGene']))
        ]
        df_VJ = df_VJ[df_VJ.bestVGene.str[:3] == df_VJ.bestJGene.str[:3]]
        df_VJ["chain"] = df_VJ.bestVGene.str[:3]
        df_VJ["VJ_pair"] = df_VJ["bestVGene"] + "_" + df_VJ["bestJGene"]
        Reads_Mapped_Confidently_to_VJ_Gene = df_VJ.shape[0]
        self.add_metric(
            name=f"{self.read_type} Mapped Confidently to VJ Gene",
            value=Reads_Mapped_Confidently_to_VJ_Gene,
            total=total_read,
        )

        # chain
        for chain in self.chains:
            df_chain = df_VJ[df_VJ.chain == chain]
            Reads_Mapped_to_chain = df_chain.shape[0]
            self.add_metric(
                name=f"{self.read_type} Mapped to {chain}",
                value=Reads_Mapped_to_chain,
                total=total_read,
            )

        # unique UMI
        df_UMI = df_VJ.drop_duplicates(subset=["barcode", "UMI"], keep="first")

        # filter1: keep top 1 in each combinations
        groupby_elements = [
            'barcode',
            'chain',
            'bestVGene',
            'bestJGene',
            'aaSeqCDR3',
            'nSeqCDR3',
        ]
        df_UMI_count = df_UMI.groupby(
            groupby_elements, as_index=False).agg({"UMI": "count"})
        df_UMI_count = df_UMI_count.sort_values("UMI", ascending=False)
        # out unfiltered
        df_UMI_count.to_csv(self.UMI_count_unfiltered_file, sep="\t", index=False)

        df_UMI_count_filter = df_UMI_count.groupby(
            ["barcode", "chain"], as_index=False).head(1)
        # out filtered
        df_UMI_count_filter.to_csv(
            self.UMI_count_filtered_file,
            sep="\t",
            index=False
        )

        if self.args.debug:
            unique_UMI = df_UMI.shape[0]
            self.add_metric(
                name="UMI unique count",
                value=unique_UMI,
                total=align_read_with_correct_CDR3,
            )

            UMI_after_Contamination_Filtering = df_UMI_count.filter.UMI.sum()
            self.add_metric(
                name="UMI after Contamination Filtering",
                value=UMI_after_Contamination_Filtering,
                total=unique_UMI,
            )

    @utils.add_log
    def fastq_to_dataframe(self):
        # read input_file
        with pysam.FastxFile(self.args.fq) as fh:
            index = 0
            read_row_list = []
            for entry in fh:
                attr = entry.name.split("_")
                barcode = attr[0]
                umi = attr[1]
                dic = {"readId": index, "barcode": barcode, "UMI": umi}
                read_row_list.append(dic)
                index += 1
            df_fastq = pd.DataFrame(read_row_list, columns=["readId", "barcode", "UMI"])
        return df_fastq

    def get_df_align(self, df_fastq):
        alignments = pd.read_csv(self.alignments, sep="\t")
        alignments.readId = alignments.readId.astype(int)
        df_fastq.readId = df_fastq.readId.astype(int)
        df_align = pd.merge(df_fastq, alignments, on="readId", how="right")
        return df_align

    def run(self):
        self.run_mixcr()
        df_fastq = self.fastq_to_dataframe()
        total_read = df_fastq.shape[0]
        df_align = self.get_df_align(df_fastq)
        self.mixcr_summary(total_read, df_align)
        self.clean_up()


@utils.add_log
def mapping_vdj(args):
    # TODO
    # add TCR or BCR prefix to distinguish them in html report summary; should improve
    step_name = f"{args.type}_mapping_vdj"
    mapping_vdj_obj = Mapping_vdj(args, step_name)
    mapping_vdj_obj.run()


def get_opts_mapping_vdj(parser, sub_program):
    parser.add_argument("--type", help='TCR or BCR', required=True)
    parser.add_argument(
        '--species',
        choices=['hs', 'mmu'],
        help='Default `hs`. `hs`(human) or `mmu`(mouse). ',
        default='hs'
    )
    parser.add_argument("--not_consensus", action='store_true', help="Input fastq is not consensused.")
    if sub_program:
        parser.add_argument(
            "--fq",
            help="Required. Input fastq file.",
            required=True,
        )
        parser = s_common(parser)



import glob
import os

import numpy as np
import pandas as pd

from celescope.tools.report import reporter
from celescope.tools.utils import add_log, format_stat, read_barcode_file


class Count_cite():

    def __init__(
        self,
        sample,
        outdir,
        assay,
        read_count_file,
        match_dir,
    ):
        self.sample = sample
        self.outdir = outdir
        self.assay = assay
        self.read_count_file = read_count_file
        self.match_dir = match_dir
        self.match_barcode, self.cell_total = read_barcode_file(match_dir)
        self.df_read_count = pd.read_csv(read_count_file, sep="\t", index_col=0)
        self.tsne_file = glob.glob(f'{match_dir}/*analysis/*tsne_coord.tsv')[0]

        if not os.path.exists(outdir):
            os.system('mkdir -p %s' % outdir)

        # out
        self.mtx = f'{outdir}/{sample}_citeseq.mtx.gz'
        self.stats = None
        self.stat_file = f'{self.outdir}/stat.txt'

    @add_log
    def run(self):
        stats = pd.Series()
        mapped_read = self.df_read_count['read_count'].sum()

        # in cell
        df_read_count_in_cell = self.df_read_count[self.df_read_count.index.isin(self.match_barcode)]
        mapped_read_in_cell = int(df_read_count_in_cell['read_count'].sum())
        stats = stats.append(pd.Series(
            format_stat(mapped_read_in_cell, mapped_read),
            index=['Mapped Reads in Cells']
        ))

        # UMI
        df_UMI_in_cell = df_read_count_in_cell.reset_index().groupby([
            'barcode', 'barcode_name']).agg({'UMI': 'count'})
        df_UMI_in_cell = df_UMI_in_cell.reset_index()
        df_UMI_in_cell = df_UMI_in_cell.pivot(
            index='barcode', columns='barcode_name', values='UMI')
        df_cell = pd.DataFrame(index=self.match_barcode)
        df_UMI_cell = pd.merge(
            df_cell,
            df_UMI_in_cell,
            how="left",
            left_index=True,
            right_index=True)

        # fillna
        df_UMI_cell.fillna(0, inplace=True)
        df_UMI_cell = df_UMI_cell.astype(int)
        df_UMI_cell_out = df_UMI_cell.T
        df_UMI_cell_out.to_csv(self.mtx, sep='\t', compression='gzip')

        # UMI
        UMIs = df_UMI_cell.apply(sum, axis=1)
        median = round(np.median(UMIs), 2)
        mean = round(np.mean(UMIs), 2)
        stats = stats.append(pd.Series(
            str(median),
            index=['Median UMI per Cell']
        ))

        stats = stats.append(pd.Series(
            str(mean),
            index=['Mean UMI per Cell']
        ))

        self.stats = stats

    def report(self):

        self.stats.to_csv(self.stat_file, sep=':', header=False)
        t = reporter(
            name='count_cite',
            assay=self.assay,
            sample=self.sample,
            stat_file=self.stat_file,
            outdir=self.outdir + '/..')
        t.get_report()



import os

import pandas as pd

import celescope.tools.utils as utils
from celescope.__init__ import __VERSION__, ASSAY_DICT
from celescope.tools.__init__ import PATTERN_DICT
from celescope.tools.barcode import Chemistry
from celescope.tools.step import Step, s_common


@utils.add_log
def sample(args):

    step_name = "sample"
    step = Step(args, step_name)

    sample_name = args.sample
    assay = args.assay
    assay_description = ASSAY_DICT[assay]
    version = __VERSION__
    outdir = args.outdir
    chemistry = args.chemistry

    # get chemistry
    if chemistry == 'auto':
        fq1 = args.fq1
        ch = Chemistry(fq1)
        chemistry = ch.check_chemistry()
        chemistry = ",".join(set(chemistry))
    else:
        chemistry = args.chemistry

    if not os.path.exists(outdir):
        os.system('mkdir -p %s' % outdir)

    stat = pd.DataFrame({
        "item": ["Sample ID", "Assay", "Chemistry", "Software Version"],
        "count": [sample_name, assay_description, chemistry, version],
    },
        columns=["item", "count"]
    )
    stat_file = outdir + "/stat.txt"
    stat.to_csv(stat_file, sep=":", header=None, index=False)

    step.clean_up()

    return chemistry


def get_opts_sample(parser, sub_program):
    if sub_program:
        parser = s_common(parser)
        parser.add_argument('--fq1', help='read1 fq file')
    parser.add_argument('--chemistry', choices=list(PATTERN_DICT.keys()), help='chemistry version', default='auto')
    return parser
